import math
import numpy as np

from ..simplicial_complexes import flag_complex as flCpx

###############################################################################
# Jump to next hypercube in cover


def next_hypercube(pos, div):
    """Jumps to next hypercube in cubical cover

    Parameters
    ----------
    pos : :obj:`list`
        List of integer values specifying the position of the current
        hypercube. This is edited to the next hypercube.

    div : :obj:`list`
        List of integer values specifying how many hypercubes divide each
        dimension.

    Example
    -------
        >>> pos = [1,1,1]
        >>> div = [3,3,2]
        >>> next_hypercube(pos, div)
        >>> pos
        [1, 2, 0]

    """
    for i in range(len(pos)):
        if pos[-i-1] < div[-i-1]-1:
            pos[-i-1] += 1
            for k in range(i):
                pos[-k-1] = 0
            break

###############################################################################
# Generate cubical cover of a point cloud


def generate_cover(max_div, overlap, point_cloud):
    """Divides a point cloud into a cubical cover.

    Receives a point cloud point_cloud in R^n and returns it divided into cubes
    and their respective intersections. It also generates the nerve of the
    covering.

    Parameters
    ----------
    max_div : int
        Number of divisions on the maximum side of point_cloud
    overlap : float
        Overlap between adjacent hypercubes.
    point_cloud : :obj:`Numpy Array`
        Each row contains the coordinates of a point

    Returns
    -------
    divided_point_cloud : :obj:`list(list(Numpy Array 2D))`
        Point cloud coordinates indexed by nerve. The `i` entry contains the
        point cloud coordinates indexed by the `i` simplices of the nerve.
        That is, the first entry contains the coordinates contained in
        hypercubes. The second entry the coordinates of points in double
        intersections of hypercubes. And so on.
    points_IN : :obj:`list(list(Numpy Array 1D))`
        Identification Numbers (IN) of points in regions indexed by nerve.
        That is, this is the same as `divided_point_cloud` but storing IN
        instead of coordinates.
    nerve : :obj:`list(Numpy Array)`
        The nerve of the hypercube cover.

    Example
    -------
        >>> point_cloud = circle(5, 1)
        >>> max_div = 2
        >>> overlap = 0.5
        >>> divided_point_cloud, points_IN, nerve = generate_cover(max_div,
        ... overlap, point_cloud)
        >>> divided_point_cloud[0]
        [array([[-0.80901699, -0.58778525],
               [ 0.30901699, -0.95105652]]), array([[ 0.30901699,  0.95105652],
               [-0.80901699,  0.58778525]]), array([[ 1.        ,  0.        ],
               [ 0.30901699, -0.95105652]]), array([[ 1.        ,  0.        ],
               [ 0.30901699,  0.95105652]])]
        >>> divided_point_cloud[1]
        [array([], shape=(0, 2), dtype=float64), array([[ 0.30901699,
        -0.95105652]]), array([], shape=(0, 2), dtype=float64), array([],
        shape=(0, 2), dtype=float64), array([[ 0.30901699,  0.95105652]]),
        array([[ 1.,  0.]])]
        >>> points_IN[0]
        [array([3, 4]), array([1, 2]), array([0, 4]), array([0, 1])]
        >>> points_IN[1]
        [array([], dtype=float64), array([4]), array([], dtype=float64),
        array([], dtype=float64), array([1]), array([0])]
        >>> nerve[0]
        4
        >>> nerve[1]
        array([[ 0.,  1.],
               [ 0.,  2.],
               [ 1.,  2.],
               [ 0.,  3.],
               [ 1.,  3.],
               [ 2.,  3.]])
        >>> nerve[2]
        array([[ 0.,  1.,  2.],
               [ 0.,  1.,  3.],
               [ 0.,  2.,  3.],
               [ 1.,  2.,  3.]])
        >>> nerve[3]
        array([[ 0.,  1.,  2.,  3.]])


    """
    # Get the dimension of our dataset
    dim = np.size(point_cloud, 1)
    # Find the 'corners' of the dataset
    min_corner, max_corner = corners_hypercube(point_cloud)
    # Find the largest coordinate difference between min_corner and max_corner
    max_length = np.amax(np.absolute(max_corner - min_corner))
    side = max_length / max_div  # side of division hypercubes
    number_hypercubes = 1
    div = []  # list containing no of divisions per dimension
    pos = []  # list containing current position
    for d in range(dim):
        div.append(math.ceil((max_corner[d] - min_corner[d]) / side))
        number_hypercubes *= div[d]  # compute total number of hypercubes
        pos.append(0)  # initialize position list to zero

    divided_point_cloud = []
    points_IN = []
    for d in range(number_hypercubes):
        divided_point_cloud.append([])
        points_IN.append([])

    """
    Now we assign points to each division hypercube.
    Also we compute the underlying graph of the nerve.
    """
    # dist: distance from the center of an hypercube to a face
    dist = overlap / 2. + side / 2.

    for i in range(number_hypercubes):
        # Find two corners and center point of small hypercube
        lcor = np.copy(min_corner)      # lower corner
        ucor = np.copy(min_corner)      # upper corner
        center = np.copy(min_corner)    # center point
        for d in range(dim):
            lcor[d] += side * pos[d]
            ucor[d] += side * pos[d]
            center[d] += side * pos[d] + side / 2.
            lcor[d] -= dist - side / 2.
            ucor[d] += dist + side / 2.

        """
        Include points in divided_point_cloud[i] if they lie on this hypercube
        """
        for idx_pt, pt in enumerate(point_cloud):
            in_hypercube = True  # assume the point lies on the hypercube
            """Now we check whether x lies really on the hypercube"""
            for k in range(dim):
                if lcor[k] > pt[k]:
                    in_hypercube = False
                    break
                if ucor[k] < pt[k]:
                    in_hypercube = False
                    break
            """If x is in hypercube we include it"""
            if in_hypercube:
                divided_point_cloud[i].append(pt)
                points_IN[i].append(idx_pt)

        divided_point_cloud[i] = np.array(divided_point_cloud[i])
        points_IN[i] = np.array(points_IN[i])
        # advance position of hypercube
        next_hypercube(pos, div)

    # Compute nerve of the cover
    nerve = nerve_hypercube_cover(div)

    # Compute the point cloud and IN for each simplex in Nerve
    # Return this information so that the function spectral_sequence
    # has less work to do. This can be parallelized
    nerve_point_cloud = [divided_point_cloud]
    nerve_points_IN = [points_IN]
    for k, k_simplices_nerve in enumerate(nerve[1:]):
        nerve_point_cloud.append([])
        nerve_points_IN.append([])
        for simplex in k_simplices_nerve:
            points_IN_intersection = intersection_covers(points_IN, simplex)
            points_coord_intersection = point_cloud[points_IN_intersection]
            nerve_points_IN[k+1].append(np.array(points_IN_intersection))
            nerve_point_cloud[k+1].append(np.array(points_coord_intersection))

    return nerve_point_cloud, nerve_points_IN, nerve


###############################################################################
# Background functions supporting generate_cover


def corners_hypercube(point_cloud):
    """Returns maximum and minimum corners of hypercube containing point_cloud.

    Parameters
    ----------
    point_cloud : :obj:`Numpy Array`
        Coordinates of point data. Each row corresponds to a point.

    Returns
    -------
    min_corner, max_corner : :obj:`Numpy Array`, :obj:`Numpy Array`
        Minimum and maximum corners of containing hypercube.

    Example
    -------
        >>> from permaviss.sample_point_clouds.examples import random_cube
        >>> point_cloud = random_cube(5,2)
        >>> point_cloud
        array([[-0.30392908, -0.40559307],
               [ 0.4736051 , -0.28257937],
               [-0.41760472, -0.30445089],
               [-0.02406966,  0.001455  ],
               [-0.28425041, -0.11212227]])
        >>> min_corner, max_corner = corners_hypercube(point_cloud)
        >>> min_corner
        array([-0.41760472, -0.40559307])
        >>> max_corner
        array([ 0.4736051,  0.001455 ])

    """
    min_corner = np.amin(point_cloud, axis=0)
    max_corner = np.amax(point_cloud, axis=0)

    return min_corner, max_corner

###############################################################################
# Generate the nerve of an hypercube covering


def nerve_hypercube_cover(div):
    """Generates the nerve of an hypercube covering.

    Given an array of divisions of an hypercube cover, this returns the nerve.

    Parameters
    ----------
    div : :obj:`Numpy Array`
        1D array of divisions per dimension.

    Returns
    -------
    nerve : :obj:`list(Numpy Array)`
        Nerve associated to hypercube covering.

    Example
    -------
    Nerve for three dimensional covering. There are 6 hypercubes and the
    divisions per dimension are given as 1 x 3 x 2.

        >>> div = [1,3,2]
        >>> nerve = nerve_hypercube_cover(div)
        >>> nerve[0]
        6
        >>> nerve[1]
        array([[ 0.,  1.],
               [ 0.,  2.],
               [ 1.,  2.],
               [ 0.,  3.],
               [ 1.,  3.],
               [ 2.,  3.],
               [ 2.,  4.],
               [ 3.,  4.],
               [ 2.,  5.],
               [ 3.,  5.],
               [ 4.,  5.]])
        >>> nerve[2]
        array([[ 0.,  1.,  2.],
               [ 0.,  1.,  3.],
               [ 0.,  2.,  3.],
               [ 1.,  2.,  3.],
               [ 2.,  3.,  4.],
               [ 2.,  3.,  5.],
               [ 2.,  4.,  5.],
               [ 3.,  4.,  5.]])
        >>> nerve[3]
        array([[ 0.,  1.,  2.,  3.],
               [ 2.,  3.,  4.,  5.]])
        >>> nerve[4]
        []

    """
    dim = np.size(div)
    # compute number of hypercubes
    number_hypercubes = 1
    for d in range(dim):
        number_hypercubes *= div[d]

    nerve_graph = []
    pos = np.zeros(dim)
    for current_index in range(number_hypercubes):
        neighbour_step = -1 * np.ones(dim)  # current step to neighbour
        # Add all the edges to the neighbour graph
        while np.any(neighbour_step < np.zeros(dim)):
            step = 1  # how much we need to move the index for each coordinate
            # Initialize and compute the index of the neighbour
            neighbour_index = current_index  # initialize the lower index
            for i in range(dim):
                # if the hypercube is not a neighbour pass
                if (
                        ((pos[i] == 0) & (neighbour_step[i] == -1)) or
                        ((pos[i] == div[i] - 1) & (neighbour_step[i] == 1))):
                    neighbour_index = -1
                    break

                neighbour_index += neighbour_step[-i - 1] * step
                step *= div[-i - 1]  # increase coordinate step

            if (neighbour_index >= 0) & (sorted(
                    [neighbour_index, current_index]) not in nerve_graph):
                """Include edge (index_neighbour,index_current) to the nerve
                graph. Only do this whenever this makes sense and the edge is
                not in the nerve already.
                """
                nerve_graph.append(sorted([neighbour_index, current_index]))

            # advance to next neighbour step
            for i in range(dim):
                if neighbour_step[-i-1] < 1:
                    neighbour_step[-i-1] += 1
                    for k in range(i):
                        neighbour_step[-k-1] = -1
                    break

        next_hypercube(pos, div)

    return flCpx.flag_complex(nerve_graph, number_hypercubes, 2**dim)

###############################################################################
# Compute the common points on the intersection determined by simplex


def intersection_covers(points_IN,  simplex):
    """Computes the points in the intersection specified by a nerve simplex.

    Parameters
    ----------
    points_IN : :obj:`list(list(Numpy Array 1D))`
        Identification Numbers (IN) of points in covering hypercubes.
    simplex : :obj:`Numpy Array`
        Simplex in nerve which specifies intersection between hypercubes.

    Returns
    -------
    points_IN_intersection : :obj:`list(int)`
        IN of points in the intersection specified by simplex.

    Example
    -------
        >>> import numpy as np
        >>> points_IN = [np.array([0,3,5]),np.array([0,1]), np.array([1,5])]
        >>> simplex = np.array([0,1])
        >>> intersection_covers(points_IN, simplex)
        [0]

    """
    points_IN_intersection = points_IN[int(simplex[0])]

    for k in range(1, len(simplex)):
        points_IN_intersection = [
                idx_pt for idx_pt in points_IN_intersection
                if idx_pt in points_IN[int(simplex[k])]]

    return points_IN_intersection



"""
    PH_classic.py

    This module implements a function which computes bases for the image and
    kernel of morphisms between persistence modules.
"""
import numpy as np

from ..gauss_mod_p import gauss_mod_p

from .barcode_bases import barcode_basis


###############################################################################
# Find pivot of array


def _pivot(l):
    """
    Compute pivot of a list of integers.

    Parameters
    ----------
    l : :obj:`list(int)`

    Returns
    -------
    index : int
        Index of last nonzero entry.
        Returns -1 if the list is zero.

    """
    l_bool = np.nonzero(l)
    if len(l_bool[0]) > 0:
        return l_bool[0][-1]

    return -1


assert _pivot(np.array([0, 1, 0, 1, 0])) == 3
assert _pivot(np.array([0, 0, 0])) == -1


###############################################################################
# Persistent homology mod p


def persistent_homology(D, R, max_rad, p):
    """
    Given the differentials of a filtered simplicial complex `X`,
    we compute its homology.

    In this function, the domain is on the columns and the range on the rows.
    Coordinates are stored as columns in an array.
    Barcode ranges are stored as pairs in an array of two columns.

    Parameters
    ----------
    D : :obj:`list(Numpy Array)`
        The ith entry stores the ith differential of the simplicial complex.
    R : :obj:`list(int)`
        The ith entry contains the radii of filtration for the ith skeleton
        of `X`. For example, the 1st entry contains, in order, the radii
        of each edge in `X`. In dimension 0 we have an empty list.
    p : int(prime)
        Chosen prime to perform arithmetic mod `p`.

    Returns
    -------
    Hom : :obj:`list(barcode_bases)`
        The `i` entry contains the `i` Persistent Homology classes for `X`.
        These are stored as :obj:`barcode_bases`. If a cycle does not die
        we put max_rad as death radius. Additionally, each entry is ordered
        according to the standard barcode order.
    Im : :obj:`list(barcode_bases)`
        The `i` entry contains the image of the `i+1` differential as
        :obj:`barcode_bases`.
    PreIm: :obj:`list(Numpy Array (len(R[*]), Im[*].dim)`
        Preimage matrices, 'how to go back from boundaries'

    Example
    -------
        >>> from permaviss.sample_point_clouds.examples import circle
        >>> from permaviss.simplicial_complexes.differentials import
        ... complex_differentials
        >>> from permaviss.simplicial_complexes.vietoris_rips import
        ... vietoris_rips
        >>> import scipy.spatial.distance as dist
        >>> point_cloud = circle(10, 1)
        >>> max_rad = 1
        >>> p = 5
        >>> max_dim = 3
        >>> Dist = dist.squareform(dist.pdist(point_cloud))
        >>> compx, R = vietoris_rips(Dist, max_rad, max_dim)
        >>> differentials = complex_differentials(compx, p)
        >>> Hom, Im, PreIm = persistent_homology(differentials, R, max_rad, p)
        >>> print(Hom[0])
        Barcode basis
        [[ 0.          1.        ]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]
         [ 0.          0.61803399]]
        [[ 1.  1.  0.  0.  0.  0.  0.  0.  0.  0.]
         [ 0.  4.  1.  0.  0.  0.  0.  0.  0.  0.]
         [ 0.  0.  4.  1.  0.  0.  0.  0.  0.  0.]
         [ 0.  0.  0.  4.  1.  0.  0.  0.  0.  0.]
         [ 0.  0.  0.  0.  4.  0.  0.  1.  0.  0.]
         [ 0.  0.  0.  0.  0.  1.  0.  4.  0.  0.]
         [ 0.  0.  0.  0.  0.  4.  0.  0.  1.  0.]
         [ 0.  0.  0.  0.  0.  0.  1.  0.  4.  0.]
         [ 0.  0.  0.  0.  0.  0.  4.  0.  0.  1.]
         [ 0.  0.  0.  0.  0.  0.  0.  0.  0.  4.]]
        >>> print(Hom[1])
        Barcode basis
        [[ 0.61803399  1.        ]]
        [[ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 4.]
         [ 1.]]
        >>> print(Im[0])
        Barcode basis
        [[ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]
         [ 0.61803399  1.        ]]
        [[ 1.  0.  0.  0.  0.  0.  0.  0.  0.]
         [ 4.  1.  0.  0.  0.  0.  0.  0.  0.]
         [ 0.  4.  1.  0.  0.  0.  0.  0.  0.]
         [ 0.  0.  4.  1.  0.  0.  0.  0.  0.]
         [ 0.  0.  0.  4.  0.  0.  1.  0.  0.]
         [ 0.  0.  0.  0.  1.  0.  4.  0.  0.]
         [ 0.  0.  0.  0.  4.  0.  0.  1.  0.]
         [ 0.  0.  0.  0.  0.  1.  0.  4.  0.]
         [ 0.  0.  0.  0.  0.  4.  0.  0.  1.]
         [ 0.  0.  0.  0.  0.  0.  0.  0.  4.]]
        >>> print(Im[1])
        Barcode basis
        []
        >>> print(PreIm[0])
        []
        >>> print(PreIm[1])
        [[ 1.  0.  0.  0.  0.  0.  0.  0.  0.]
         [ 0.  1.  0.  0.  0.  0.  0.  0.  0.]
         [ 0.  0.  1.  0.  0.  0.  0.  0.  0.]
         [ 0.  0.  0.  1.  0.  0.  0.  0.  0.]
         [ 0.  0.  0.  0.  1.  0.  0.  0.  0.]
         [ 0.  0.  0.  0.  0.  1.  0.  0.  0.]
         [ 0.  0.  0.  0.  0.  0.  1.  0.  0.]
         [ 0.  0.  0.  0.  0.  0.  0.  1.  0.]
         [ 0.  0.  0.  0.  0.  0.  0.  0.  1.]
         [ 0.  0.  0.  0.  0.  0.  0.  0.  0.]]

    """
    dim = len(D)

    # Introduce list of Hom_bars, Hom_coord, Im_bars, Im_coord, and Preim
    Hom_bars = []
    Hom_coord = []
    Im_bars = []
    Im_coord = []
    PreIm = []
    for d in range(dim):
        Hom_bars.append([])
        Hom_coord.append([])
        Im_bars.append([])
        Im_coord.append([])
        PreIm.append([])

    # preallocate space for dim - 1
    domain_dim = np.size(D[dim - 1], 1)
    Hom_bars[dim - 1] = np.zeros((domain_dim, 2))
    Hom_coord[dim - 1] = np.zeros((domain_dim, domain_dim))
    Hom_dim_next = 0
    # Perform Gaussian eliminations, starting from D[dim-1] and ending on D[1].
    pivots = []
    for d in range(dim - 1, 0, -1):
        # Compute dimensions for domain and range of D[d]
        domain_dim = np.size(D[d], 1)
        range_dim = np.size(D[d], 0)
        Daux = np.copy(D[d])
        # Set pivot columns automatically to zero (clear optimization)
        Daux[:, pivots] = np.zeros((range_dim, len(pivots)))
        # Compute pivots complement and reset pivots to zero
        non_pivots = np.setdiff1d(range(domain_dim), pivots)
        pivots = []
        # preallocate space for the dimension d - 1
        Hom_bars[d - 1] = np.zeros((range_dim, 2))
        Hom_coord[d - 1] = np.zeros((range_dim, range_dim))
        Im_bars[d - 1] = np.zeros((len(non_pivots), 2))
        Im_coord[d - 1] = np.zeros((range_dim, len(non_pivots)))
        PreIm[d] = np.zeros((domain_dim, len(non_pivots)))
        # Perform Gaussian reduction by left to right column additions mod p
        Im_aux, T = gauss_mod_p.gauss_col(Daux, p)
        # Reset dimension variables
        Hom_dim_current = Hom_dim_next
        Hom_dim_next = 0
        Im_dim_next = 0
        # Go through all reduced columns of Im_aux which are not pivots
        for k in non_pivots:
            reduced_im = Im_aux[:, k]
            # If column is nonzero
            if np.any(reduced_im):
                pivots.append(_pivot(reduced_im))
                birth_rad = R[d - 1][_pivot(reduced_im)]
                death_rad = R[d][k]
                if birth_rad < death_rad:
                    Hom_bars[d - 1][Hom_dim_next] = [birth_rad, death_rad]
                    Hom_coord[d - 1][:, Hom_dim_next] = reduced_im
                    Hom_dim_next += 1
                # end if

                Im_bars[d - 1][Im_dim_next] = [death_rad, max_rad]
                Im_coord[d - 1][:, Im_dim_next] = reduced_im
                PreIm[d][:, Im_dim_next] = T[:, k]
                Im_dim_next += 1
            # end if
            # If column is zero
            else:
                birth_rad = R[d][k]
                death_rad = max_rad
                Hom_bars[d][Hom_dim_current] = [birth_rad, death_rad]
                Hom_coord[d][:, Hom_dim_current] = T[:, k]
                Hom_dim_current += 1
            # end else
        # end for

        # Get rid of extra preallocated storage
        Hom_bars[d] = Hom_bars[d][:Hom_dim_current]
        Hom_coord[d] = Hom_coord[d][:, :Hom_dim_current]
        Im_bars[d - 1] = Im_bars[d - 1][:Im_dim_next]
        Im_coord[d - 1] = Im_coord[d - 1][:, :Im_dim_next]
        PreIm[d] = PreIm[d][:, :Im_dim_next]
    # end for

    # Get infinite barcodes of Z[0]
    Hom_dim_current = Hom_dim_next
    non_pivots = np.setdiff1d(range(D[0]), pivots)
    for k in non_pivots:
        Hom_bars[0][Hom_dim_current] = [0, max_rad]
        Hom_coord[0][k, Hom_dim_current] = 1
        Hom_dim_current += 1
    # end for

    # free extra preallocated space
    Hom_bars[0] = Hom_bars[0][:Hom_dim_current]
    Hom_coord[0] = Hom_coord[0][:, :Hom_dim_current]

    # Store as persistence bases
    Hom = []
    Im = []
    # Reference to corresponding dimension R.
    # Here, we create barcode basis for the underlying complexes,
    # the case dim=0 is special.
    for d in range(dim):
        barcode_simplices = np.append([R[d]], [max_rad * np.ones(len(R[d]))],
                                      axis=0).T
        basis = barcode_basis(barcode_simplices)
        if d == 0:
            basis.dim = D[0]
        # end if
        if np.size(Hom_bars[d], 0) > 0:
            Hom.append(barcode_basis(Hom_bars[d], basis, Hom_coord[d]))
            Hom[d].sort()
        else:
            Hom.append(barcode_basis([]))
        # end else
        if len(Im_bars[d]) > 0:
            # Order PreIm according to order of Im
            Im.append(barcode_basis(Im_bars[d], basis, Im_coord[d],
                                    store_well_defined=True))
            PreIm[d+1] = PreIm[d+1][:, Im[-1].well_defined]
            order = Im[d].sort(send_order=True)
            PreIm[d+1] = PreIm[d+1][:, order]
        else:
            Im.append(barcode_basis([]))
        # end else
    # end for

    """Return everything."""
    return Hom, Im, PreIm



###############################################################################
# flag complexes for graphs:
# This is used e.g. for computing the nerve of a cover.

import numpy as np

###############################################################################


def _lower_neighbours(G, u):
    """Given a graph `G` and a vertex `u` in `G`, we return a list with the
    vertices in `G` that are lower than `u` and are connected to `u` by an
    edge in `G`.

    Parameters
    ----------
    G : :obj:`Numpy Array(no. of edges, 2)`
        Matrix storing the edges of the graph.
    u : int
        Vertex of the graph.

    Returns
    -------
    lower_neighbours : :obj:`list`
        List of lower neighbours of `u` in `G`.
    """
    lower_neighbours = []
    for e in G:
        if max(e) == u:
            lower_neighbours.append(min(e))

    return np.unique(lower_neighbours)

###############################################################################
# Main function


def flag_complex(G, no_vertices, max_dim):
    """Compute the flag complex of a graph `G` up to a maximum dimension `max_dim`.

    Parameters
    ----------
    G : :obj:`Numpy Array(no. of edges, 2)`
        Matrix storing the edges of the graph.
    no_vertices : int
        Number of vertices in graph `G`
    max_dim : int
        Maximum dimension

    Returns
    -------
    fl_cpx : :obj:`list(Numpy Array)`
        Flag complex of `G`. The `0` entry stores the number of vertices. For
        a higher entry `i`, `fl_cpx[i]` stores a :obj:`Numpy Array` matrix
        with the `i` simplices from `fl_cpx`.

    """
    if max_dim < 0:
        print("Cannot compute a complex of dimension: {}".format(max_dim))
        raise ValueError

    fl_cpx = []
    for i in range(max_dim + 1):
        fl_cpx.append([])

    fl_cpx[0] = no_vertices

    if max_dim == 0:
        return fl_cpx

    fl_cpx[1] = np.copy(G)

    # Build flag complex inductively
    for d in range(1, max_dim):
        for simplex in fl_cpx[d]:
            N = _lower_neighbours(fl_cpx[1], simplex[0])
            for v in simplex[1:]:
                N = np.intersect1d(N, _lower_neighbours(fl_cpx[1], v))

            # find simplices containing simplex and add them
            if np.size(N) > 0:
                simplices = np.ones((np.size(N), 1 + np.size(simplex)))
                simplices = np.multiply(np.append([1], simplex), simplices)
                simplices[:, 0] = N
                if isinstance(fl_cpx[d+1], list):
                    fl_cpx[d + 1] = simplices
                else:
                    fl_cpx[d+1] = np.append(fl_cpx[d + 1], simplices, 0)

    return fl_cpx



"""
    barcode_bases.py

"""
import numpy as np


class barcode_basis(object):
    """This class implements barcode bases.

    Associated bars and coordinates are given for some barcode base.
    Each coordinate is stored in a column, whereas bars are stored on rows.
    This generates a barcode basis with all the input data.
    There is the exception of broken barcode bases, which come up
    when solving the extension problem.

    Note
    ----
    Barcode bases are assumed to be well defined in the sense that:

        1)They are linearly independent with respect to boxplus operation

        2)They generate the respective module or submodule.


    Parameters
    ----------
    bars : :obj:`Numpy Array (dim, 2)`
        Each entry is a pair specifying birth and death radius of a bar.
    prev_basis : reference to a previously defined :class:`barcode_basis`.
        This will be the basis in which the coordinates are given.
    coordinates : :obj:`Numpy Array (dim, prev_basis.dim)`
        Cooridnates of this basis in terms of `prev_basis`
    store_well_defined : bool, default is `False`
        Whether we want to store the indices of well defined bars.
        That is, whether we wish to store indices of bars where the
        birth radius is strictly smaller than the death radius.
    broken_basis : bool, default is `False`
        Whether the barcode basis is `broken`. This appears when solving
        the extension problem. A barcode base is `broken` if it is not natural.
    broken_differentials : :obj:`Numpy Array (dim, dim)`
        Matrix of broken differentials.
        These give coefficients of a barcode generator in term of other
        generators. This is used when the given generator dies, and we
        write it in terms of other generators that are still alive.

    Returns
    -------
    :obj:`barcode_basis`

    Raises
    ------
    ValueError
        If `prev_coord.dim` is different to the number of rows in `coordinates`

    ValueError
        If the number of rows in `bar` is different to the number of columns
        in `coordinates`.

    ValueError
        If `broken_basis = True` but `broken_differentials` is not given.

    Examples
    --------

        >>> import numpy as np
        >>> bars = np.array([[0,2],[0,1],[2,3],[2,2]])
        >>> base1 = barcode_basis(bars, store_well_defined=True)
        >>> base1.dim
        3
        >>> base1.well_defined
        array([ True,  True,  True, False], dtype=bool)
        >>> print(base1)
        Barcode basis
        [[0 2]
         [0 1]
         [2 3]]


        >>> bars = np.array([[0,2],[2,3]])
        >>> coordinates = np.array([[1,0],
        ...                         [1,0],
        ...                         [0,2]])
        >>> base2 = barcode_basis(bars, prev_basis, coordinates)
        >>> base2.dim
        2
        >>> print(base2)
        Barcode basis
        [[0 2]
         [2 3]]
        [[1 0]
         [1 0]
         [0 2]]

        >>> bars = np.array([[0,2],[0,1],[1,3]])
        >>> broken_differential = np.array(
        ... [[0,0,0],
        ...  [0,0,0],
        ...  [0,1,0]])
        >>> base3 = barcode_basis(bars, broken_basis=True,
        ... broken_differentials=broken_differential)
        >>> base3.dim
        3
        >>> print(base3)
        Barcode basis
        [[0 2]
         [0 1]
         [1 3]]


    """
    def __init__(self, bars, prev_basis=None, coordinates=np.array([[]]),
                 store_well_defined=False, broken_basis=False,
                 broken_differentials=None):
        """Constructor method
        """
        # When the basis is broken, the death_differentials must be given
        if broken_basis and (broken_differentials is None):
            raise ValueError

        self.broken_basis = broken_basis

        # Store given data
        self.barcode = np.copy(bars)
        self.prev_basis = prev_basis
        self.coordinates = coordinates
        # Assume the barcode basis is not sorted
        self.sorted = False
        # Take out trivial bars and also bad defined bars
        if len(bars) == 0:
            self.dim = 0
        else:
            well_defined = self.barcode[:, 0] < self.barcode[:, 1]
            self.barcode = self.barcode[well_defined]
            self.dim = np.sum(well_defined)

        # Add coordinates, checking for mistakes on the data.
        if (prev_basis is not None) and (np.size(coordinates, 1) > 0):
            # Check that coordinates match the dimension of the previous basis
            if np.size(coordinates, 0) != self.prev_basis.dim:
                print("Error: given coordinates do not match dimension of" +
                      "prev_basis")
                raise ValueError

            # check that same amount of coordinates as bars were given
            if np.size(bars, 0) != np.size(coordinates, 1):
                print("Error:given {} bars but {} coordinates".format(
                        len(bars), np.size(coordinates, 1)))
                raise ValueError

            self.coordinates = self.coordinates[:, well_defined]

        if store_well_defined:
            self.well_defined = well_defined

        # make sure the broken differentials are given as a square
        # matrix of dimension self.dim
        if broken_basis:
            if np.size(broken_differentials, 0) == self.dim and np.size(
                    broken_differentials, 1) == self.dim:
                self.broken_differentials = broken_differentials
            else:
                raise ValueError

    def __str__(self):
        """Printing function for barcode bases.
        """
        print("Barcode basis")
        print(self.barcode)
        if self.prev_basis is not None:
            print(self.coordinates)

        return ""

    def sort(self, precision=7, send_order=False):
        """Sorts a barcode basis according to the standard barcode order.

        That is, from smaller birth radius to bigger, and from
        bigger death radius to smaller. A precision up to n zeros is an
        optional argument.

        Parameters
        ----------
        precision : int, default is 7
            Number of zeros of precision.
        send_order : bool, default is False
            Whether we want to generate a :obj:`Numpy Array` storing
            the original order.

        Returns
        -------
        :obj:`Numpy Array`
            One dimensional array storing original order of barcodes.

        Example
        -------

            >>> bars = np.array([[1,2],[0,4],[-1,3],[1,4]])
            >>> base4 = barcode_basis(bars)
            >>> base4.sort(send_order=True)
            array([2, 1, 3, 0])
            >>> base4 = barcode_basis(bars)
            >>> print(base4)
            Barcode basis
            [[ 1  2]
             [ 0  4]
             [-1  3]
             [ 1  4]]
            >>> base4.sort(send_order=True)
            array([2, 1, 3, 0])
            >>> print(base4)
            Barcode basis
            [[-1  3]
             [ 0  4]
             [ 1  4]
             [ 1  2]]

        """

        aux_bars = np.around(self.barcode, decimals=precision)
        order = np.lexsort((-aux_bars[:, 1], aux_bars[:, 0]))
        if self.prev_basis is not None:
            self.coordinates = self.coordinates[:, order]

        self.barcode = self.barcode[order]
        self.sorted = True
        if self.broken_basis:
            self.broken_differentials = self.broken_differentials[order]
            self.broken_differentials = self.broken_differentials[:, order]

        if send_order:
            return order

    ###########################################################################
    # Functions used by Image_kernel:

    def changes_list(self):
        """Returns an array with the values of changes occurring in the basis.

        Returns
        -------
        :obj:`Numpy Array`
            One-dimensional array with radii where either a bar dies or is
            born in `self`.

        Example
        -------
            >>> print(base1)
            Barcode basis
            [[0 2]
             [0 1]
             [2 3]]
            >>> base1.changes_list()
            array([0, 1, 2, 3])

        """
        changes_radii = np.resize(self.barcode, (1, 2 * len(self.barcode)))
        return np.unique(changes_radii)

    def active(self, rad, start=0, end=-1):
        """Returns array with active indices at rad.

        Option to restrict to generators from start to self.dim

        Parameters
        ----------
        rad : `float`
            Radius where we want to check which barcodes are active.
        start : int, default is 0
            Generator in `self` from which we start to search.
        end : int, default is -1
            Generator in `self` until which we end searching for active
            generators.

        Returns
        -------
        :obj:`Numpy Array`
            One-dimensional array with indices of active generators.
            This is relative to the start index.

        Example
        -------
            >>> print(base3)
            Barcode basis
            [[0 2]
             [0 1]
             [1 3]]
            >>> base3.active(1.2)
            array([0, 2])
            >>> base3.active(0.8, start=1)
            array([0])


        """
        if end == -1:
            end = self.dim
        # end if
        if start >= self.dim or (end > self.dim) or (end < 0):
            print("start:{}, end:{}, self.dim:{}".format(start, end, self.dim))
            raise ValueError
        # end if
        indices = np.array(range(end - start))
        active_bool = np.logical_and(self.barcode[:, 0] <= rad,
                                     rad < self.barcode[:, 1])
        return indices[active_bool[start:end]]

    def trans_active_coord(self, coord, rad, start=0):
        """Given coordinates on the active generators, this returns
        coordinates in the whole basis.

        This also can be done relative to a start index higher than 0

        Parameters
        ----------
        coord : :obj:`Numpy Array`
            One dimensional array specifying the coordinates on the active
            basis.
        rad : float
            Radius on which we are checking for active generators.
        start : int, default is 0
            We ignore the indices smaller than this. Notice that `coord`
            as well as the produced absolute coordinates will be adjusted
            appropriately.

        Returns
        -------
        :obj:`Numpy Array`
            One dimensional array with a

        Example
        -------
            >>> bars = [[0,5],[0,2],[1,4],[0.5,3]]
            >>> base7 = barcode_basis(bars)
            >>> base7.trans_active_coord([0,1,1],2.4)
            array([ 0.,  0.,  1.,  1.])
            >>> base7.trans_active_coord([-1,1],3.8)
            array([-1.,  0.,  1.,  0.])
            >>> base7.trans_active_coord([-1,1],2.2,1)
            array([ 0., -1.,  1.])


        """
        trans_coord = np.zeros(self.dim - start)
        trans_coord[self.active(rad, start)] = coord
        return trans_coord

    def death(self, rad, start=0):
        """Returns an array with the indices dying at rad.

        These indices are relative to the optional argument start.

        Parameters
        ----------
        rad : float
            Radius at which generators might be dying
        start : int, default is 0


        Example
        -------
            >>> print(base4)
            Barcode basis
            [[-1  3]
             [ 0  4]
             [ 1  4]
             [ 1  2]]
            >>> base4.death(4)
            array([1, 2])
            >>> base4.death(4,2)
            array([0])

        """
        if start >= self.dim:
            raise ValueError

        indices = np.array(range(self.dim - start))
        death_bool = (self.barcode[:, 1] == rad)
        return indices[death_bool[start:]]

    ###########################################################################
    # Functions used by spectral_sequence class

    def birth_radius(self, coordinates):
        """This finds the birth radius of a list of coordinates.

        Example
        -------
            >>> base3 = barcode_basis([[0,2],[0,1],[1,3]])
            >>> base3.birth_radius([1,0,3])
            1


        """
        if not np.any(coordinates):
            return np.nan  # Returns nan if coordinates are zero

        return np.max(self.barcode[np.nonzero(coordinates), 0])

    def death_radius(self, coordinates):
        """Find the death radius of given coordinates.

        Example
        -------
            >>> print(base3)
            Barcode basis
            [[0 2]
             [0 1]
             [1 3]]
            >>> base3.death_radius([1,1,1])
            3


        """
        if not np.any(coordinates):
            return np.nan  # Returns nan if coordinates are zero

        return np.max(self.barcode[np.nonzero(coordinates)])

    def bool_select(self, selection):
        """Given a boolean array, we select generators from a barcode basis.

        Example
        -------
            >>> print(base3)
            Barcode basis
            [[0 2]
             [0 1]
             [1 3]]
            >>> base5 = base3.bool_select([True,False,True])
            >>> print(base5)
            Barcode basis
            [[0 2]
             [1 3]]


        """
        if self.prev_basis is None:
            return barcode_basis(self.barcode[selection])

        return barcode_basis(self.barcode[selection], self.prev_basis,
                             self.coordinates[:, selection])

    def active_coordinates(self, rad):
        """Active submatrix of coordinates at rad

        Example
        -------
            >>> bars = np.array(
            ... [[0,5],[1,5],[1,5],[2,4]])
            >>> baseP = barcode_basis(bars)
            >>> barsC = np.array(
            ... [[1,4],[2,5]])
            >>> coord = np.array(
            ... [[1,0],
            ...  [0,1],
            ...  [2,1],
            ...  [0,1]])
            >>> baseC = barcode_basis(barsC, baseP, coord)
            >>> print(baseC)
            Barcode basis
            [[1 4]
             [2 5]]
            [[1 0]
             [0 1]
             [2 1]
             [0 1]]
            >>> bars
            array([[0, 5],
                   [1, 5],
                   [1, 5],
                   [2, 4]])
            >>> baseC.active_coordinates(1.3)
            array([[1],
                   [0],
                   [2]])
            >>> baseC.active_coordinates(4.8)
            array([[0],
                   [1],
                   [1]])
            >>> baseC.active_coordinates(3)
            array([[1, 0],
                   [0, 1],
                   [2, 1],
                   [0, 1]])

        """
        A = np.copy(self.coordinates[:, self.active(rad)])
        return A[self.prev_basis.active(rad)]

    def active_domain(self, rad):
        """Active columns of coordinates at rad

        Example
        -------

            >>> print(baseC)
                Barcode basis
                [[1 4]
                 [2 5]]
                [[1 0]
                 [0 1]
                 [2 1]
                 [0 1]]
            >>> baseC.active_domain(1.3)
            array([[1],
                   [0],
                   [2],
                   [0]])

        """
        if self.dim == 0:
            return np.array([])
        else:
            return np.copy(self.coordinates[:, self.active(rad)])

    ###########################################################################
    # Function updating broken_differentials

    def update_broken(self, A, rad):
        """Updates a matrix A, using the broken differentials of the
        generators dying at rad.

        We assume that the broken barcode basis is indexing the rows of A.

        Parameters
        ----------
        A : :obj:`Numpy Array`
            columns represent coordinates in the barcode_basis.
        rad : float
            Radius at which we want to update the coordinates using the
            broken_differentials.

        Returns
        -------
        A : :obj:`Numpy Array`
            return updated matrix

        Example
        -------
            >>> print(base3)
            Barcode basis
            [[0 2]
             [0 1]
             [1 3]]
            >>> A = np.array(
            ... [[1,1,1],
            ...  [0,2,-1],
            ...  [0,1,1]])
            >>> base3.update_broken(A, 1)
            array([[1, 1, 1],
                   [0, 0, 0],
                   [0, 3, 0]])


        """
        dying_indices = self.death(rad)
        if len(dying_indices) > 0:
            sum_update = np.matmul(self.broken_differentials[:, dying_indices],
                                   A[dying_indices])
            A[dying_indices] = np.zeros((len(dying_indices), np.size(A, 1)))
            A = A + sum_update

        return A



"""
    image_kernel.py

    This module implements a function which computes bases for the image and
    kernel of morphisms between persistence modules.
"""
import numpy as np

from .barcode_bases import barcode_basis

from ..gauss_mod_p import gauss_mod_p
from ..gauss_mod_p.functions import multiply_mod_p

###############################################################################
# Find pivot of array


def _pivot(l):
    """
    Parameters
    ----------
    l : :obj:`list(int)`
        List of integers to compute pivot from.

    Returns
    -------
    int
        Index of last nonzero entry on `l`. Returns -1 if the list is zero.
    """
    l_bool = np.nonzero(l)
    if len(l_bool[0]) > 0:
        return l_bool[0][-1]

    return -1


assert _pivot(np.array([0, 1, 0, 1, 0])) == 3
assert _pivot(np.array([0, 0, 0])) == -1


###############################################################################
# Find barcode generators for image and kernel of a persistence morphism.

def image_kernel(A, B, F, p, start_index=0, prev_basis=None):
    """
    This computes basis for the image and kernel of a persistence morphism.
        f: A --> B

    This is the algorithm described in https://arxiv.org/abs/1907.05228.
    Recall that for such an algorithm to work the `A` and `B` must be ordered.
    This is why the function first orders the barcode generators from
    start_index until A.dim.
    Additionally, it also orders the barcodes from B. By 'ordered' we mean that
    the barcodes are sorted according to the standard order of barcodes.

    It can also compute relative barcode bases for the image. This is used when
    computing quotients. The optional argument start_index indicates the
    minimum index from which we want to compute barcodes relative to the
    previous generators. That is, given start_index, the function will return
    image barcodes for

            <F[start_dim, ..., A.dim]>  mod <F[0,1,..., start_dim-1]>.

    At the end the bases for the image and kernel are
    returned in terms of the original ordering.

    Additionally, this handles the case for when B is a broken barcode basis.
    Notice that in such a case, only the barcode basis of the image will be
    computed

    Parameters
    ----------
    A : :class:`barcode_basis` object
        Basis of domain.
    B : :class:`barcode_basis` object
        Basis of range.  This can be a `broken` barcode basis.
    F : Numpy Array (`B`.dim, `A`.dim)
        Matrix associated to the considered persistence morphism.
    p : int
        prime number of finite field.
    start_index : int, default is 0
        Index from which we get a barcode basis for the image.
    prev_basis : int, default is None
        If  `start_index` > 0, we need to also give a reference to a basis
        of barcodes from A[start_dim] until A[A.dim].

    Returns
    -------
    Ker : :class:`barcode_basis` object
        Absolute/relative basis of kernel of f.
    Im : :class:`barcode_basis` object
        Absolute/relative basis of image of f.
    PreIm : Numpy Array (A.dim, Im.dim)
        Absolute/relative preimage coordinates of f. That is, each column
        stores the sums that generate the corresponding Image barcode.

    Examples
    --------

        >>> import numpy as np
        >>> from permaviss.persistence_algebra.barcode_bases import
        ... barcode_basis
        >>> A = barcode_basis([[1,8],[1,5],[2,5], [4,8]])
        >>> B = barcode_basis([[-1,3],[0,4],[0,3.5],[2,5],[2,4],[3,8]])
        >>> F = np.array([[4,1,1,0],[1,4,1,0],[1,1,4,0],[0,0,1,4],[0,0,4,1],
        ... [0,0,0,1]])
        >>> p = 5
        >>> Im, Ker, PreIm = image_kernel(A,B,F,p)
        >>> print(Im)
        Barcode basis
        [[ 1.   4. ]
         [ 1.   3.5]
         [ 2.   5. ]
         [ 4.   8. ]]
        [[ 4.  0.  1.  0.]
         [ 1.  0.  1.  0.]
         [ 1.  2.  4.  0.]
         [ 0.  0.  1.  4.]
         [ 0.  0.  4.  1.]
         [ 0.  0.  0.  1.]]
        >>> print(Ker)
        Barcode basis
        [[ 3.5  8. ]
         [ 4.   5. ]]
        [[ 1.  0.]
         [ 1.  4.]
         [ 0.  0.]
         [ 0.  0.]]
        >>> print(PreIm)
        [[ 1.  1.  0.  0.]
         [ 0.  1.  0.  0.]
         [ 0.  0.  1.  0.]
         [ 0.  0.  0.  1.]]

    Note
    ----
    This algorithm will only work if the matrix of the persistence morphism
    is well defined. That is, a generator can only map to generators that have
    been born and have not yet died.

    """
    # If the basis B is not ordered, we order it
    B_origin = B
    return_order_B = range(B.dim)
    if not B.sorted:
        B = barcode_basis(B.barcode)
        order = B.sort(send_order=True)
        F = np.copy(F[order])
        return_order_B = np.argsort(order)

    # Order A from start_index until A.dim
    A_origin = A
    return_order_A = range(A.dim)
    if not A.sorted:
        A_rel = barcode_basis(A.barcode[start_index:])
        order = A_rel.sort(send_order=True)
        shifted_order = order + start_index
        shifted_order = np.append(range(start_index), shifted_order,
                                  axis=0).astype(int)
        return_order_A = np.argsort(order)
        F = np.copy(F[:, shifted_order])
        A = barcode_basis(np.append(A.barcode[:start_index], A_rel.barcode,
                                    axis=0))

    # Get sorted radii where changes happen
    a = np.sort(np.unique(np.append(A.changes_list(), B.changes_list())))
    rel_A_dim = A.dim - start_index
    # Save space for Im barcode and coordinates
    Im_barcode = np.concatenate(
        ([A.barcode[start_index:, 0]], [np.zeros(rel_A_dim)]),
        axis=0).T
    Im_rel_coordinates = np.copy(F[:, start_index:])
    # Save space for Ker barcode and coordinates
    Ker_barcode = np.zeros((rel_A_dim, 2))
    Ker_coordinates = np.zeros((rel_A_dim, rel_A_dim))
    kernel_dim = 0
    # Save space for preimage
    PreIm = np.identity(rel_A_dim)
    # Tracking variables for deaths of image and kernel barcodes
    # These store indices from 0 to rel_A_dim
    dead_images = []
    dead_kernels = []
    # Initialize a copy of F which will be changing as we iterate over a
    Im_coordinates = np.copy(F)
    # Go through all radius where a change happens
    for i, rad in enumerate(a):
        if B.broken_basis:
            Im_coordinates = B.update_broken(Im_coordinates, rad)

        # Update I0 from submatrix of Im
        I0 = Im_coordinates[B.active(rad)][:, A.active(rad)]
        # Get the active indices in A and the relative active indices in A
        active_indices = A.active(rad)
        active_rel_indices = A.active(rad, start=start_index)
        # Take care of special case when I0 = []
        if len(I0) == 0:
            # Find active domain elements that have not died yet
            active_not_dead = np.setdiff1d(active_rel_indices, dead_images)
            for j in active_not_dead:
                Im_barcode[j, 1] = rad
                Ker_barcode[kernel_dim, 0] = rad
                Ker_coordinates[j, kernel_dim] = 1
                kernel_dim += 1
                dead_images.append(j)
            # end for
        # end if

        # Find dying domain generators whose image has not died yet
        dying_generators = A.death(rad, start=start_index)
        alive_dying = np.setdiff1d(dying_generators, dead_images)
        # Set the death radius for the image barcodes
        for j in alive_dying:
            Im_barcode[j, 1] = rad
            dead_images.append(j)
        # end for

        if kernel_dim > 0 and not B.broken_basis:
            # Update K0 from submatrix of Ker
            K0 = Ker_coordinates[:, :kernel_dim]
            K0 = K0[active_rel_indices]
            if len(K0) == 0:
                dying_kernels = np.setdiff1d(range(kernel_dim), dead_kernels)
                for j in dying_kernels:
                    dead_kernels.append(j)
                    Ker_barcode[j, 1] = rad
                # end for
            # end if
            else:
                # Reduce K0 by column additions
                K0, Q0 = gauss_mod_p.gauss_col(K0, p)
                # Perform the same reductions on Ker, if Q0 is not empty
                if np.any(Q0):
                    Ker_coordinates[:, :kernel_dim] = multiply_mod_p(
                        Ker_coordinates[:, :kernel_dim], Q0, p)
                # Compute the start_index of active barcodes
                start_index_active = len(active_indices) - len(
                    active_rel_indices)
                # Look at K0, eliminating linear dependencies and killing
                # barcodes on image.
                for j, c in enumerate(K0.T):
                    # Find pivot in terms of whole basis A, keep also the
                    # active pivot.
                    active_pivot = _pivot(c)
                    piv = active_rel_indices[active_pivot]
                    # If the pivot is new and the column is nonzero
                    if (piv not in dead_images) and (active_pivot > -1):
                        # Add new coordinates to Image, set barcode endpoint,
                        # store preimage.
                        A_rel_coord = A.trans_active_coord(
                            c, rad, start=start_index)
                        image_A_rel = multiply_mod_p(
                            F[:, start_index:], np.transpose([A_rel_coord]),
                            p)[:, 0]
                        Im_rel_coordinates[:, piv] = image_A_rel
                        Im_coordinates[
                            :, start_index + piv] = np.copy(image_A_rel)
                        Im_barcode[piv, 1] = rad
                        PreIm[:, piv] = A_rel_coord
                        # Set corresponding column in I0 to zero and add piv
                        # to dead images
                        I0[
                            :, active_pivot + start_index_active
                            ] = np.zeros(np.size(I0, 0))
                        dead_images.append(piv)
                    # end if
                    # If a barcode is dying in the kernel, add death radius
                    # to Ker_barcode
                    if (active_pivot == -1) and (j not in dead_kernels):
                        dead_kernels.append(j)
                        Ker_barcode[j, 1] = rad
                    # end if
                # end for
            # end else
        # end if
        # Reduce I0, adding new generators to the kernel and adjusting
        # Im_rel_coordinates.
        if I0.size > 0:
            I0, T0 = gauss_mod_p.gauss_col(I0, p)
            # Adapt T0 to the whole basis A
            Q0 = np.zeros((A.dim, np.size(T0, 1)))
            Q0[active_indices] = T0
            Q = np.identity(A.dim)
            Q[:, active_indices] = Q0
            # Perform same additions in Im_coordinates
            Im_coordinates = multiply_mod_p(Im_coordinates, Q, p)
            # Then, adapt Q to the relative basis of A
            Q = Q[start_index:, start_index:]
            # Then perform the additions in PreIm and Im
            PreIm = multiply_mod_p(PreIm, Q, p)
            Im_rel_coordinates = multiply_mod_p(Im_rel_coordinates, Q, p)
            # Now, we check for 0 columns in I0, iterating over
            # active_indices >= start_index.
            rel_I0_T = I0.T[active_indices >= start_index]
            start_kernel_dim = kernel_dim
            for j, c in enumerate(rel_I0_T):
                piv = active_rel_indices[j]
                if (not np.any(c)) and (piv not in dead_images):
                    Ker_coordinates[:, kernel_dim] = PreIm[:, piv]
                    Ker_barcode[kernel_dim, 0] = rad
                    kernel_dim += 1
                    Im_barcode[piv, 1] = rad
                    dead_images.append(piv)
                # end if
            # end for
            # Reduce new additions to kernel
            if (kernel_dim > start_kernel_dim) and not B.broken_basis:
                # Update K0 from submatrix of Ker
                K0 = Ker_coordinates[:, :kernel_dim]
                K0 = K0[active_rel_indices]
                # Reduce K0 by column additions
                K0, Q0 = gauss_mod_p.gauss_col(K0, p)
                # Perform the same reductions on Ker, if Q0 is not empty
                if np.any(Q0):
                    Ker_coordinates[
                        :, :kernel_dim
                        ] = multiply_mod_p(Ker_coordinates[:, :kernel_dim],
                                           Q0, p)
                # Look at K0, eliminating linear dependencies
                for j, c in enumerate(K0.T):
                    if (_pivot(c) == -1) and (j not in dead_kernels):
                        dead_kernels.append(j)
                        Ker_barcode[j, 1] = rad
                    # end if
                # end for
            # end if
        # end if
        # Go to next radius in a

    # end for

    # Store the barcode endpoints for the image and kernel generators that are
    # still alive.
    for j in range(rel_A_dim):
        if j not in dead_images:
            Im_barcode[j, 1] = a[-1]
        # end if
    # end for
    for j in range(kernel_dim):
        if j not in dead_kernels:
            Ker_barcode[j, 1] = a[-1]
        # end if
    # end for
    # Return to normal order in B
    Im_rel_coordinates = Im_rel_coordinates[return_order_B]
    # Return to normal order in A
    PreIm = PreIm[return_order_A]
    Ker_coordinates = Ker_coordinates[return_order_A]
    # Create barcode basis for image, and adjust PreIm according to well
    # defined barcodes.
    Im_basis = barcode_basis(Im_barcode, B_origin, Im_rel_coordinates,
                             store_well_defined=True)
    PreIm = PreIm[:, Im_basis.well_defined]
    # Order Im and store the order in PreIm
    order = Im_basis.sort(send_order=True)
    PreIm = PreIm[:, order]
    # If the basis of B is broken, return Im_basis
    if B.broken_basis:
        return Im_basis

    # Create a barcode basis for the kernel and sort it.
    if start_index > 0:
        Ker_basis = barcode_basis(Ker_barcode[:kernel_dim], prev_basis,
                                  Ker_coordinates[:, :kernel_dim])
    else:
        Ker_basis = barcode_basis(Ker_barcode[:kernel_dim], A_origin,
                                  Ker_coordinates[:, :kernel_dim])

    Ker_basis.sort()

    return Im_basis, Ker_basis, PreIm



""" __init__.py

This module implements the Gaussian elimination of a matrix by column
reductions.  Coefficients are assumed to lie on a finite field Z (mod p) Where
p is a prime and Z are the integers.  Ideally, this should be implemented in
Cython or similar later on.
"""



"""
This module implements the Mayer-Vietoris spectral sequence management.
"""

import numpy as np
import scipy.spatial.distance as dist

from multiprocessing import Pool
from functools import partial

from ..covers import cubical_cover

from ..simplicial_complexes.differentials import complex_differentials
from ..simplicial_complexes.vietoris_rips import vietoris_rips

from ..persistence_algebra.PH_classic import persistent_homology
from ..persistence_algebra.image_kernel import image_kernel
from ..persistence_algebra.module_persistence_homology import (
     module_persistence_homology)
from ..persistence_algebra.barcode_bases import barcode_basis

from .spectral_sequence_class import spectral_sequence


def create_MV_ss(point_cloud, max_r, max_dim, max_div, overlap, p):
    """ This function creates a Mayer Vietoris spectral sequence with the
    given parameters. The procedure has four main steps:

    1) Obtain a cover and a nerve associated to it.

    2) Compute the persistent homology on each cover, intersections, and so on.

    3) Compute spectral sequence pages until they collapse.

    4) Solve the extension problem

    Parameters
    ----------
    point_cloud : Numpy Array
        Coordinates for given points. Each row corresponds to a point.
    max_r : float
        Maximum radius of persistence.
    max_dim : int
        Maximum dimension of simplexes in Vietoris-Rips complex.
    max_div : int
        Number of division hypercubes on the dimension with maximum length on
        point cloud.
    overlap : float
        Overlap between adjacent covers.

    Returns
    -------
    MV_ss : :class:`spectral_sequence` object containing all the information.

    Example
    -------
        >>> from permaviss.sample_point_clouds.examples import random_cube,
        ... take_sample
        >>> X = random_cube(1000,3)
        >>> point_cloud = take_sample(X,130)
        >>> max_r = 0.36
        >>> max_dim = 3
        >>> p = 3
        >>> max_div = 2
        >>> overlap = max_r*1.01
        >>> MV_ss = create_MV_ss(point_cloud, max_r, max_dim, max_div, overlap,
        ... p)
        PAGE: 1
        [[ 25   7   4   1   0   0   0   0   0]
         [160 118 128 144 112  56  16   2   0]
         [310 380 436 445 336 168  48   6   0]]
        PAGE: 2
        [[ 21   0   0   0   0   0   0   0   0]
         [ 98   2   0   0   0   0   0   0   0]
         [131   5   1   0   0   0   0   0   0]]
        PAGE: 3
        [[ 21   0   0   0   0   0   0   0   0]
         [ 97   2   0   0   0   0   0   0   0]
         [131   5   0   0   0   0   0   0   0]]
        PAGE: 4
        [[ 21   0   0   0   0   0   0   0   0]
         [ 97   2   0   0   0   0   0   0   0]
         [131   5   0   0   0   0   0   0   0]]
        >>> print(MV_ss.persistent_homology[0].dim)
        131
        >>> print(MV_ss.persistent_homology[1].dim)
        97
        >>> print(MV_ss.persistent_homology[2].dim)
        21

    """
    # Divide point cloud using hypercube cover
    # Use points_IN to build Nerve in future more general version
    nerve_point_cloud, points_IN, nerve = cubical_cover.generate_cover(
        max_div, overlap, point_cloud)

    nerve_dim = len(nerve)
    # Count maximum points on hypercube cover
    max_points = 0
    for hyp_pc in nerve_point_cloud[0]:
        if max_points < np.size(hyp_pc, 0):
            max_points = np.size(hyp_pc, 0)

    # initialize the spectral sequence, compute the maximum number of pages
    no_pages = min(max_dim + 2, nerve_dim)
    MV_ss = spectral_sequence(nerve, nerve_point_cloud, points_IN, max_dim,
                              max_r, no_pages, p)

    # 0 PAGE

    for n_dim in range(0, nerve_dim):
        if n_dim > 0:
            n_spx_number = np.size(nerve[n_dim], 0)
        else:
            n_spx_number = nerve[0]

        partial_persistent_homology = partial(local_persistent_homology,
                                              nerve_point_cloud,
                                              max_r, max_dim, p, n_dim)

        workers_pool = Pool()
        output = workers_pool.map(partial_persistent_homology,
                                  range(n_spx_number))
        workers_pool.close()

        MV_ss.add_output_first(output, n_dim)

    # PAGES => 1
    for current_page in range(1, no_pages):
        # Print page
        print("PAGE: {}".format(current_page))
        flip = np.array(range(MV_ss.no_rows))
        flip = -flip
        print(MV_ss.page_dim_matrix[current_page][np.argsort(flip)])
        # Loop through sequences of differentials ending in the first
        # current_page columns
        for start_n_dim in range(current_page):
            for start_deg in range(max_dim):
                base = []
                differentials = [MV_ss.page_dim_matrix[
                    current_page, start_deg, start_n_dim]]
                deg = start_deg
                n_dim = start_n_dim
                # While (n_dim, deg) lies within the spectral sequence
                # boundaries
                while deg >= 0 and n_dim < nerve_dim:
                    # Generate barcode bases for sending to
                    # module_persistence_homology
                    if current_page == 1:
                        if MV_ss.page_dim_matrix[1][deg][n_dim] == 0:
                            barcode = []
                        else:
                            barcode = np.zeros((MV_ss.page_dim_matrix[
                                1][deg][n_dim], 2))
                            prev_cycle_dim = 0
                            for nerv_spx_index, cycle_dim in enumerate(
                                    MV_ss.cycle_dimensions[n_dim][deg]):
                                if cycle_dim - prev_cycle_dim > 0:
                                    barcode[prev_cycle_dim: cycle_dim] = \
                                        MV_ss.Hom[0][n_dim][nerv_spx_index][
                                            deg].barcode
                                # end if
                                prev_cycle_dim = cycle_dim
                            # end for
                        # end else
                        base.append(barcode_basis(barcode))
                    else:
                        if type(MV_ss.Hom[current_page - 1][n_dim][
                                deg]) == list:
                            MV_ss.Hom[current_page - 1][n_dim][
                                deg] = barcode_basis([])
                        # end if
                        base.append(MV_ss.Hom[current_page - 1][n_dim][deg])
                    # end else
                    # Advance to differential domain and compute differential
                    deg += 1 - current_page
                    n_dim += current_page
                    if deg >= 0 and n_dim < nerve_dim:
                        differentials.append(MV_ss.zig_zag(n_dim, deg,
                                                           current_page).T)
                    # end if
                # end while
                Hom, Im, PreIm = module_persistence_homology(differentials,
                                                             base, p)
                MV_ss.add_output_higher(Hom, Im, PreIm, start_n_dim, start_deg,
                                        current_page)
            # end for
        # end for
    # end for

    # compute representatives for cycles
    for n_dim in range(MV_ss.no_columns):
        for deg in range(MV_ss.no_rows):
            if MV_ss.page_dim_matrix[no_pages-1, deg, n_dim] > 0:
                _ = MV_ss.zig_zag(n_dim, deg, n_dim + deg + 2, store_reps=True)
                if(MV_ss.page_dim_matrix[no_pages-1, deg, n_dim] != len(
                        MV_ss.tot_complex_reps[n_dim][deg][0])):
                    raise ValueError
            # end if
        # end for
    # end for

    # EXTENSION PROBLEM
    MV_ss.persistent_homology.append(MV_ss.Hom[no_pages-1][0][0])
    # Go through each diagonal
    for deg in range(1, MV_ss.no_rows):
        # compute dimension of diagonal deg
        dim_PH = [0]
        # Cumulative dimensions for persistent homology along diagonals
        start_deg = deg
        for start_n_dim in range(min(deg+1, MV_ss.no_columns)):
            dim_PH.append(MV_ss.page_dim_matrix[no_pages][
                start_deg, start_n_dim] + dim_PH[-1])
            start_deg -= 1
        # end for
        if dim_PH[-1] > 0:
            # Save space for extension matrix
            ext_mat = np.zeros((dim_PH[-1], dim_PH[-1]))
            # compute extension matrix
            start_deg = deg-1
            for start_n_dim in range(1, min(deg+1, MV_ss.no_columns)):
                if MV_ss.page_dim_matrix[no_pages-1][
                        start_deg, start_n_dim] > 0:
                    extensions = MV_ss.extension(start_n_dim, start_deg)
                    column_range = ext_mat[:, dim_PH[start_n_dim]: dim_PH[
                        start_n_dim+1]]
                    for i, blk in enumerate(extensions):
                        column_range[dim_PH[start_n_dim - i]: dim_PH[
                            start_n_dim + 1 - i]] = extensions[i]
                    # end for
                # end if
                start_deg -= 1
            # end for
            #
            # add up barcodes in diagonal
            barcode = []
            start_deg = deg
            for start_n_dim in range(min(deg+1, MV_ss.no_columns)):
                if len(barcode) == 0 and \
                        MV_ss.Hom[no_pages-1][start_n_dim][start_deg].dim > 0:
                    barcode = MV_ss.Hom[no_pages-1][start_n_dim][
                        start_deg].barcode
                else:
                    if MV_ss.Hom[no_pages-1][start_n_dim][start_deg].dim > 0:
                        barcode = np.append(barcode, MV_ss.Hom[no_pages-1][
                            start_n_dim][start_deg].barcode, axis=0)
                # end else
                start_deg -= 1
            # end for
            # Create a new barcode basis, based on the death radius of the
            # extension matrix column
            diagonal_basis = barcode_basis(np.copy(barcode))
            # Create new basis, which is the direct sum matrix. This is a
            # broken barcode basis with broken differentials given by the
            # extension matrix.
            direct_sum_basis = barcode_basis(barcode, broken_basis=True,
                                             broken_differentials=ext_mat)
            # Create an identity matrix as a morphism from diagonal_basis
            # to direct_sum_basis.
            diag_diff = np.identity(diagonal_basis.dim)
            # Order the barcodes for domain and range
            MV_ss.order_diagonal_basis.append(diagonal_basis.sort(
                                              send_order=True))
            order = direct_sum_basis.sort(send_order=True)
            # Reorder associated matrix accordingly
            diag_diff = diag_diff[:, MV_ss.order_diagonal_basis[-1]]
            diag_diff = diag_diff[order]
            # set all barcode endpoints to max_r in diagonal_basis
            diagonal_basis.barcode[:, 1] = max_r * np.ones(diagonal_basis.dim)
            PH = image_kernel(diagonal_basis, direct_sum_basis,
                              diag_diff, MV_ss.p)
            MV_ss.persistent_homology.append(PH)
        # end if
    # end for
    return MV_ss

###############################################################################
# Local persistent homology
# Method to be parallelized


def local_persistent_homology(nerve_point_cloud, max_r, max_dim, p,  n_dim,
                              spx_idx):
    """ This function computes the Vietoris Rips complex and persistent
    homology of a covering region.

    It is meant to be run in parallel.

    Parameters
    ----------
    nerve_point_cloud : :obj:`list(list(Numpy Array))`
        Local point cloud coordinates indexed by nerve. The first entry
        contains a list of the point cloud coordinates for each covering
        region. The second entry contains a list of the point cloud coordinates
        for each double intersection of covering regions. And so on.
    points_IN : :obj:`list(list(Numpy Array))`
        Local Identification Numbers (IN) indexed by nerve. That is, this is
        the same as `nerve_point_cloud`, but containing IN instead of
        coordinates for each point.
    max_r : `float`
        Maximum radius for computing persistent homology.
    max_dim : int
        Maximum dimension for complexes
    n_dim : int
        Current dimension in Nerve of cover
    spx_idx : int
        Index of n_dim simplex of the covering nerve.

    Returns
    -------
    local_complex : :obj:`list(Numpy Array)`
        See :mod:`permaviss.simplicial_complexes.vietoris_rips`
    local_differentials : :obj:`list(Numpy Array)`
        See :mod:`permaviss.simplicial_complexes.differentials`
    Hom, Im, PreIm : :obj:`list(barcode_basis)`, :obj:`list(barcode_basis)`,
                     :obj:`list(Numpy Array)`
        See
        :meth:`permaviss.persistence_algebra.PH_classic.persistent_homology`

    """
    local_point_cloud = nerve_point_cloud[n_dim][spx_idx]
    # Compute local Vietoris Rips complex and differentials
    if len(local_point_cloud) == 0:
        local_Dist = []
    else:
        local_Dist = dist.squareform(dist.pdist(local_point_cloud))

    local_complex, local_R = vietoris_rips(local_Dist, max_r, max_dim)
    local_differentials = complex_differentials(local_complex, p)

    # Persistent Homology
    Hom, Im, PreIm = persistent_homology(local_differentials, local_R,
                                         max_r, p)

    return local_complex, local_differentials, Hom, Im, PreIm



"""
    functions.py

    This code implements multiplication mod p and solving a linear
    equation mod p.
"""
import numpy as np

from .gauss_mod_p import gauss_col

###############################################################################
# Multiply two matrices mod p


def multiply_mod_p(A, B, p):
    """
        Multiply matrices mod p.
    """
    return np.matmul(A, B) % p


###############################################################################
# solve_mod_p
# Function to solve linear equations mod p

def solve_mod_p(A, b, p):
    """
    Find the vector x such that A * x = b (mod p)

    This method assumes that a solution exists to the equation
    A * x = b (mod p). If a solution does not exist, it raises a ValueError
    exception.

    Parameters
    ----------
    A : :obj:`Numpy Array`
        2D array
    b : :obj:`Numpy Array`
        1D array
    p : int(prime)
        Number to mod out by.

    Returns
    -------
    x : :obj:`Numpy Array`
        1D array. Solution to equation.

    Raises
    ------
    ValueError
        If a solution to the equation does not exist.
    """
    R, T = gauss_col(np.append(A, np.array([b]).T, axis=1), p)
    if np.any(R[:, -1]):
        print("Linear equation has no solution.")
        raise ValueError

    # Return last column from T without the last row
    return -T[:, -1][:-1] % p


###############################################################################
# solve_matrix_mod_p
#

def solve_matrix_mod_p(A, B, p):
    """
    Same as :meth:`solve_mod_p`, but with B and X being matrices.

    That is, given two matrices A and B, we want to find a matrix X
    such that A * X = B (mod p)

    Parameters
    ----------
    A : :obj:`Numpy Array`
        2D array
    B : :obj:`Numpy Array`
        2D array
    p : int(prime)

    Returns
    -------
    X : :obj:`Numpy Array`
        2D array solution.

    Raises
    ------
    ValueError
        There is no solution to the given equation
    """
    R, T = gauss_col(np.append(A, B.T, axis=1), p)
    if np.any(R[:, np.size(A, 1):]):
        print("Linear matrix equation has no solution.")
        raise ValueError

    # Return matrix X
    return - T[:, np.size(A, 1):][:np.size(A, 1)] % p



"""
    vietoris_rips.py
"""

import numpy as np


def _lower_neighbours(G, u):
    """ Given a graph `G` and a vertex `u` in `G`, we return a list with the
    vertices in `G` that are lower than `u` and are connected to `u` by an
    edge in `G`.

    Parameters
    ----------
    G : :obj:`Numpy Array(no. of edges, 2)`
        Matrix storing the edges of the graph.
    u : int
        Vertex of the graph.

    Returns
    -------
    lower_neighbours : :obj:`list`
        List of lower neighbours of `u` in `G`.
    """
    lower_neighbours = []  # list of lowest neighbours to be computed.
    for e in G:
        if e[1] == u:
            lower_neighbours.append(e[0])

    lower_neighbours.sort()
    return lower_neighbours


def vietoris_rips(Dist, max_r, max_dim):
    """This computes the Vietoris-Rips complex with simplexes of dimension less
    or equal to max_dim, and with maximum radius specified by max_r

    Parameters
    ----------
    Dist : :obj:`Numpy Array(no. of points, no. of points)`
        Distance matrix of points.
    max_r : float
        Maximum radius of filtration.
    max_dim : int
        Maximum dimension of computed Rips complex.

    Returns
    -------
    C : :obj:`list(Numpy Array)`
        Vietoris Rips complex generated for the given parameters. List where
        the first entry stores the number of vertices, and all other entries
        contain a :obj:`Numpy Array` with the list of simplices in `C`.
    R : :obj:`list(Numpy Array)`
        List with radius of birth for the simplices in `C`. The `i` entry
        contains a 1D :obj:`Numpy Array` containing each of the birth radii
        for each `i` simplex in `C`.

    """
    if max_dim < 1:  # at least returns always the neighbourhood graph
        max_dim = 1

    C = []
    R = []
    for i in range(max_dim + 1):
        C.append([])
        R.append([])

    # The zero component contains the number of vertices of C
    C[0] = len(Dist)
    R[0] = np.zeros(len(Dist))
    # Start with C[1]
    for i in range(C[0]):
        for j in range(i):
            if Dist[i][j] < max_r:
                C[1].append([j, i])
                R[1].append(Dist[i][j])

    # Sort edges according to their birth radius
    sort_R = np.argsort(R[1])
    R[1], C[1] = np.array(R[1]), np.array(C[1])
    R[1] = R[1][sort_R]
    C[1] = C[1][sort_R]

    # Build VR-complex inductively
    for d in range(1, max_dim):
        # Assume that C[d] is already defined, then we compute C[d+1]
        for k, s in enumerate(C[d]):
            """Find neighbouring vertices of s in C, such that they are smaller than
            all the vertices from s. """
            low_n = _lower_neighbours(C[1], s[0])
            for v in s[1:]:
                low_n = [n for n in low_n if
                         n in _lower_neighbours(C[1], v)]

            for n in low_n:
                simplex_rad = R[d][k]
                for v in s:
                    if Dist[v][n] > simplex_rad:
                        simplex_rad = Dist[v][n]

                C[d + 1].append(np.insert(s, 0, n))
                R[d + 1].append(simplex_rad)

        # Sort simplices according to their birth radius
        sort_R = np.argsort(R[d + 1])
        R[d + 1], C[d + 1] = np.array(R[d + 1]), np.array(C[d + 1])
        R[d + 1] = R[d + 1][sort_R]
        C[d + 1] = C[d + 1][sort_R]

    # store complexes as integers
    for c in C[1:]:
        c = c.astype(int)

    return C, R



"""
arithmetic_mod_c.py

Arithmetic functions for working mod c
Also includes function for inverses mod a prime number p
"""
import numpy as np


def add_mod_c(a, b, c):
    """
    Integer addition mod c.

    Parameters
    ----------

    a,b : int
        Integers to be added
    c : int
        Integer to mod out by

    Returns
    -------
    s : a + b (mod c)
    """
    s = a + b
    return s % c


def add_arrays_mod_c(A, B, c):
    """
    Adds two arrays mod c.

    Parameters
    ----------
    a,b : :obj:`Numpy Array(int)`
        Two integer arrays to be added.
    c : int
        Integer to mod out by.

    Returns
    -------
    C : :obj:`Numpy Array(int)`
        a + b (mod c)

    Raises
    ------
    ValueError
        If `len(a)` != `len(b)`

    """
    if len(A) != len(B):
        print("Trying to add two arrays with different lengths!")
        raise ValueError

    N = len(A)
    C = np.zeros(N)
    # C = (A + B) % np.array(c)
    for i in range(N):
        C[i] = add_mod_c(A[i], B[i], c)

    return C


def inv_mod_p(a, p):
    """
    Returns the inverse of a mod p

    Parameters
    ----------
    a : int
        Number to compute the inverse mod p
    p : int(prime)

    Returns
    -------
    m : int
        Integer such that m * a = 1 (mod p)

    Raises
    ------
    ValueError
        If p is not a prime number
    """
    x0, x1 = 0, 1
    y0, y1 = 1, 0
    b = p
    while a != 0:
        q, b, a = b // a, a, b % a
        y0, y1 = y1, y0 - q * y1
        x0, x1 = x1, x0 - q * x1

    if b != 1:
        print("Error:{} is not a prime number".format(p))
        raise ValueError

    if x0 < 0:
        x0 = x0 + p
    return x0



"""
    module_persistence_homology.py

    This module implements the persistence module homology
"""
import numpy as np

from .barcode_bases import barcode_basis

from .image_kernel import image_kernel

###############################################################################
# Find pivot of array


def _pivot(l):
    """Given a 1D array of integers, return the index of the last nonzero entry.

    Parameters
    ----------
    l : :obj:`list`
        1D array of integers.

    Returns
    -------
    int
        Index of last nonzero entry. If `l` is zero, returns -1.

    """
    l_bool = np.nonzero(l)
    if len(l_bool[0]) > 0:
        return l_bool[0][-1]

    return -1


assert _pivot(np.array([0, 1, 0, 1, 0])) == 3
assert _pivot(np.array([0, 0, 0])) == -1


###############################################################################
# Quotient of barcode bases


def quotient(M, N, p):
    """Assuming that N generates a submodule of M, we compute
    a barcode basis for the quotient M / N.

    Parameters
    ----------
    M : :obj:`barcode_basis`
        Basis for module

    N : :obj:`barcode_basis`
        Basis for submodule of N

    p : int(prime)

    Returns
    -------
    Q : :obj:`barcode_basis`
        Barcode basis for the quotient M / N

    """
    if N.dim == 0:
        return M

    # Check that the barcode bases coincide
    if M.prev_basis != N.prev_basis:
        raise ValueError

    # We create N_M, which is not a basis, but will be used as the domain
    # basis in image_kernel so that the quotient mod N can be performed
    N_M_barcode = np.concatenate((N.barcode, M.barcode), axis=0)
    N_M = barcode_basis(N_M_barcode)
    # create a matrix (N|M) in M.prev_basis
    matrix_N_M = np.concatenate((N.coordinates, M.coordinates), axis=1)
    # Compute quotient barcodes for M mod N
    Q, Ker, PreIm = image_kernel(N_M, M.prev_basis, matrix_N_M, p,
                                 start_index=N.dim, prev_basis=M)
    return Q


###############################################################################
# Persistence module homology mod p


def module_persistence_homology(D, Base, p):
    """Given the differentials of a chain of tame persistence modules, we
    compute barcode bases for the homology of the chain.

    Parameters
    ----------
    D : :obj:`list(Numpy Array)`
        List of differentials of the chain complex.
    Base : :obj:`Numpy Array`
        List containing barcode bases for each dimension
    p : int(prime)
        Prime number to perform arithmetic mod p

    Returns
    -------
    Hom : :obj:`list(barcode_basis)`
        Cycles mod boundaries of differentials, starting with: birth rad,
        death rad. If a cycle does not die we put max_rad as death radius.
    Im : :obj:`list(barcode_basis)`
        List storing bases for the images of differentials
    PreIm : :obj:`list(Numpy Array)`
        List storing bases for the preimages of the differentials. That is,
        which generators produce each image generator. This leads to how to
        `go back` from boundaries to preimages.

    """
    dim = len(Base)
    # lists of homology, boundaries and preboundaries
    Hom = []
    Im = []
    PreIm = []
    for d in range(dim):
        Hom.append([])
        Im.append([])
        PreIm.append([])

    Im[dim - 1] = barcode_basis([], Base[d])
    # Find bases for Kernel and Image, starting from D[dim-1], ending on D[1].
    for d in range(dim - 1, 0, -1):
        # Handle trivial cases
        if Base[d].dim == 0:
            Im[d-1] = barcode_basis([], Base[d-1])
            Hom[d] = barcode_basis([], Base[d])
        elif Base[d-1].dim == 0:
            Im[d-1] = Base[d-1]
            Hom[d] = Base[d]
        else:
            # compute barcode bases for image and kernel
            Im[d - 1], Ker, PreIm[d] = image_kernel(Base[d], Base[d-1], D[d],
                                                    p)
            # Perform quotient of Ker by Im[d]
            Hom[d] = quotient(Ker, Im[d], p)

    # Trivial case dim 0
    if Base[0].dim == 0:
        Hom[0] = Base[0]
    else:
        # All the generators from Base[0] lie in the kernel, since D[0]=0
        Ker = barcode_basis(Base[0].barcode, Base[0], np.identity(Base[0].dim))
        Ker.sort()
        if Ker.prev_basis != Im[0].prev_basis:
            print(Ker.prev_basis)
            print(Im[0].prev_basis)
            assert False

        Hom[0] = quotient(Ker, Im[0], p)

    # Return lists of barcode bases for Hom, Im and PreIm
    return Hom, Im, PreIm



"""gauss_mod_p.py

This module implements Gaussian elimination by columns modulo a prime
number p.
"""
import numpy as np
from .arithmetic_mod_p import add_arrays_mod_c, inv_mod_p

###############################################################################
# Index searching function


def _index_pivot(l):
    """Returns the pivot of a 1D array

    Parameters
    ----------
    l : :obj:`list(int)`
        List of integers to compute pivot from.

    Returns
    -------
    int
        Index of last nonzero entry on `l`. Returns -1 if the list is zero.
    """
    l_bool = np.nonzero(l)
    if len(l_bool[0]) > 0:
        return l_bool[0][-1]

    return -1


assert _index_pivot(np.array([0, 1, 0, 1, 0])) == 3
assert _index_pivot(np.array([0, 0, 0])) == -1

###############################################################################
# Gaussian elimination procedure


def gauss_col(A, p):
    """This function implements the Gaussian elimination by columns.

    A is reduced by left to right column additions. The reduced matrix has
    unique column pivots.

    Parameters
    ----------
    A : :obj:`Numpy Array`
        Matrix to be reduced
    p : `int(prime)`
        Prime number. The corresponding field will be Z mod p.

    Returns
    -------
    R : :obj:`Numpy Array`
        Reduced matrix by left to right column additions.
    T : :obj:`Numpy Array`
        Matrix recording additions performed, so that AT = R
    """
    if np.size(A, 0) == 0:
        return np.array([]), np.array([])

    # number of columns in A
    N = np.size(A, 1)
    # copy of matrix to be reduced
    # The matrix is transposed for more computational efficiency
    R = np.copy(np.transpose(A))
    T = np.identity(N)
    # iterate over all columns
    for j in range(N):
        pivot = _index_pivot(R[j])
        # Assume that the j-column is not reduced
        reduced = False
        while (pivot > -1) & (not reduced):
            reduced = True
            # look for previous columns to j
            for k in range(j):
                # if the pivots coincide, subtract column k to column j
                # multiplied by a suitable coefficient q
                if _index_pivot(R[k]) == pivot:
                    q = (R[j][pivot] * inv_mod_p(R[k][pivot], p)) % p
                    R[j] = add_arrays_mod_c(R[j], -q * R[k], p)
                    T[j] = add_arrays_mod_c(T[j], -q * T[k], p)
                    # reset pivot
                    if np.any(R[j]):
                        pivot = _index_pivot(R[j])
                        reduced = False

                    break
                # end if
            # end for
        # end while
    # end for
    return np.transpose(R), np.transpose(T)



import numpy as np

import scipy.spatial.distance as dist


def take_sample(point_cloud, no_samples):
    """ Take a subsample from samples using a minmax algorithm.

    We start from a random point. Then choose the point further appart. Next,
    we take the point that is further appart from the taken points. Continuing
    we take all the samples from `point_cloud`.

    Parameters
    ----------
    point_cloud : :obj:`Numpy Array`
        List of points to take samples from.
    np_samples : int
        Number of samples that we want to take. It has to be smaller than
        the dimension of `point_cloud`.

    Returns
    -------
    :obj:`Numpy Array`
        Matrix storing the coordinates of the sampled points.

    """
    no_points = len(point_cloud)
    no_samples = min(no_points, no_samples)
    Dist = dist.squareform(dist.pdist(point_cloud))
    k = int(np.random.rand() * no_points) - 1
    selection = [k]
    point_selection = [point_cloud[k]]
    for i in range(no_samples):
        not_selected = np.setdiff1d(range(no_points), selection)
        rest_Dist = np.min(Dist[not_selected][:, selection], axis=1)
        k = not_selected[np.argmax(rest_Dist)]
        selection.append(k)
        point_selection.append(point_cloud[k])

    return np.array(point_selection)


def circle(no_points, radius):
    """ Take random points from a circle on the plane. This circle has
    centre `(0, 0)`.

    Parameters
    ----------
    no_points : int
        Number of points wanted.
    radius : float
        Radius of the circle.

    Returns
    -------
    :obj:`Numpy Array`
        Coordinates of sampled points from the circle.

    """
    point_cloud = []
    step = 2 * np.pi / (no_points)
    for n in range(no_points):
        point_cloud.append([radius * np.cos(n * step),
                            radius * np.sin(n * step)])

    return np.asarray(point_cloud)


def random_circle(no_points, radius, epsilon, center=[0, 0]):
    """ Take random points around a circle on the plane.

    Parameters
    ----------
    no_points : int
        Number of points wanted.
    radius : float
        Radius of the circle.
    epsilon : float
        Noise that we want to apply to each sampled point.
    centre : `list(float, float)`
        Two entries specifying the position of the centre.

    Returns
    -------
    :obj:`Numpy Array`
        Coordinates of sampled points from around the circle.

    """
    point_cloud = []
    for n in range(no_points):
        random_angle = np.random.rand() * 2 * np.pi
        random_rad = radius * (1 + np.random.rand() * epsilon)
        point_cloud.append([center[0] + random_rad * np.cos(random_angle),
                            center[1] + random_rad * np.sin(random_angle)])

    return np.asarray(point_cloud)


def ball(no_points, radius, dim):
    """ Take random points from a ball of a given dimension. This ball has
    centre `(0, 0, 0)`.

    Parameters
    ----------
    no_points : int
        Number of points wanted.
    radius : float
        Radius of the ball.
    dim : int
        Dimension of the ball.

    Returns
    -------
    :obj:`Numpy Array`
        Coordinates of sampled points from the ball.

    """
    point_cloud = []
    for n in range(no_points):
        valid_point = False
        while valid_point is False:
            point = []
            for d in range(dim):
                point.append((np.random.rand() - 0.5) * radius)

            v = np.array(point)
            if np.linalg.norm(v) < radius:
                valid_point = True
                point_cloud.append(point)

    return np.asarray(point_cloud)


def random_sphere(no_points, radius, dim):
    """ Take random points from a sphere of a given dimension. This sphere has
    centre `(0, 0, 0)`.

    Parameters
    ----------
    no_points : int
        Number of points wanted.
    radius : float
        Radius of the sphere.
    dim : int
        Dimension of the sphere.

    Returns
    -------
    :obj:`Numpy Array`
        Coordinates of sampled points from the sphere.

    """
    point_cloud = []
    for n in range(no_points):
        point = []
        for d in range(dim):
            point.append((np.random.rand() - 0.5) * radius)

        v = np.array(point)
        norm = np.linalg.norm(v)
        for d in range(dim):
            point[d] = point[d] / norm

        point_cloud.append(point)

    return np.asarray(point_cloud)


def random_cube(no_points, dim):
    """ Take random points from a unit cube around the origin. This cube can be
    of various dimensions.

    Parameters
    ----------
    no_points : int
        Number of points wanted.
    dim : int
        Dimension of the cube.

    Returns
    -------
    :obj:`Numpy Array`
        Coordinates of sampled points from the cube.

    """
    point_cloud = []
    for n in range(no_points):
        point = []
        for d in range(dim):
            point.append(np.random.rand() - 0.5)

        point_cloud.append(point)

    return np.asarray(point_cloud)


def grid(hdiv, vdiv):
    """ Take the nodes from a hdiv x vdiv grid on 2D plane.

    Parameters
    ----------
    hdiv : int
        Number of rows.
    vdiv : int
        Number of columns.

    Returns
    -------
    :obj:`Numpy Array`
        List of points.

    """
    hshift = hdiv / 2. - 0.5
    vshift = vdiv / 2. - 0.5
    point_cloud = []
    for i in range(hdiv):
        for j in range(vdiv):
            point_cloud.append([i - hshift, j - vshift])

    return np.asarray(point_cloud)


def grid_tridimensional(hdiv, vdiv, ddiv):
    """ Take the nodes from a hdiv x vdiv x ddiv grid on 3D space.

    Parameters
    ----------
    hdiv : int
        Number of rows.
    vdiv : int
        Number of columns.
    ddiv : int
        Number of flats.

    Returns
    -------
    :obj:`Numpy Array`
        List of points.

    """
    hshift = hdiv / 2. - 0.5
    vshift = vdiv / 2. - 0.5
    dshift = ddiv / 2. - 0.5

    point_cloud = []
    for i in range(hdiv):
        for j in range(vdiv):
            for k in range(ddiv):
                point_cloud.append([i - hshift, j - vshift, k - dshift])

    return np.asarray(point_cloud)


def torus3D(no_points, min_rad=1, max_rad=3):
    """ Take samples from a torus embedded in 3D space.

    Parameters
    ----------
    no_points : int
        Number of points to be taken.
    min_rad : float, default is `1`
        Radius of circle on section of the torus.
    max_rad : float, default is `3`
        Distance from the torus centre to the centre of the section.

    Returns
    -------
    :obj:`Numpy Array`
        List of points.

    """
    if min_rad > max_rad:
        raise ValueError
    point_cloud = []
    for i in range(no_points):
        theta = np.random.rand() * 2 * np.pi
        alpha = np.random.rand() * 2 * np.pi
        point_cloud.append([
            (max_rad + min_rad * np.cos(alpha)) * np.cos(theta),
            (max_rad + min_rad * np.cos(alpha)) * np.sin(theta),
            min_rad * np.sin(alpha)])
    return np.asarray(point_cloud)


def torus(div, min_rad, max_rad):
    """ Take samples from a torus in 4D space.

    Parameters
    ----------
    no_points : int
        Number of points to be taken.
    min_rad : float, default is `1`
        Radius of circle on section of the torus.
    max_rad : float, default is `3`
        Distance from the torus centre to the centre of the section.

    Returns
    -------
    :obj:`Numpy Array`
        List of points.

    """
    point_cloud = []
    step_small = 2 * np.pi / div
    step_big = step_small * min_rad / max_rad
    ang_max = 0

    while(ang_max < 2 * np.pi):

        ang_min = 0

        while(ang_min < 2 * np.pi):

            point_cloud.append([min_rad * np.cos(ang_min),
                                min_rad * np.sin(ang_min),
                                max_rad * np.cos(ang_max),
                                max_rad * np.sin(ang_max)])
            ang_min += step_small

        ang_max += step_big
    return np.asarray(point_cloud)



""" Setup script. """

from setuptools import setup, find_packages

from src.permaviss.version import __version__

with open("README.rst", "r") as readme_file:
    README = readme_file.read()

setup(
    name="permaviss",
    version=__version__,
    description="Persistence Mayer Vietoris spectral sequence",
    long_description=README,
    long_description_content_type="text/x-rst",
    author_email="atorras1618@gmail.com",
    author="Alvaro Torras Casas",
    url="https://github.com/atorras1618/PerMaViss",
    download_url="https://github.com/atorras1618/PerMaViss/tarball/v0.0.1",
    licence="MIT",
    keywords=["spectral sequence", "persistent homology", "Mayer Vietoris"],
    packages=find_packages("src"),
    package_dir={"": "src"},
)



#
#    spectral_sequence_class.py
#

import numpy as np

from ..simplicial_complexes.differentials import complex_differentials
from ..gauss_mod_p.functions import solve_mod_p, multiply_mod_p


class spectral_sequence(object):
    """Space and methods for Mayer-Vietoris spectral sequences

    Parameters
    ----------
    nerve : :obj:`list(Numpy Array)`
        Simplicial complex storing the nerve of the covering. This is stored as
        a list, where the ith entry contains a Numpy Array storing all the
        ith simplices.
    nerve_point_cloud : :obj:`list(list(Numpy Array))`
        Point clouds indexed by nerve of the cover, see
        :mod:`permaviss.covers.cubical_cover`
    points_IN : :obj:`list(list(Numpy Array))`
        Point Identification Numbers (IN) indexed by nerve of the cover, see
        :mod:`permaviss.covers.cubical_cover`
    max_dim : int
        Maximum dimension of simplices.
    max_r : float
        Maximum persistence radius.
    no_pages : int
        Number of pages of the spectral sequence
    p : int(prime)
        The prime number so that our computations are mod p

    Attributes
    ----------
    nerve, nerve_point_cloud, points_IN, max_dim, max_r, no_pages, p :
        as described above
    no_rows, no_columns : int, int
        Number of rows and columns in each page
    nerve_differentials : :obj:`list(Numpy Array)`
        List storing the differentials of the Nerve. The ith entry stores the
        matrix of the ith differential.
    subcomplexes : :obj:`list(list(list(Numpy Array)))`
        List storing the simplicial complex on each cover element. For integers
        `n_dim`, `k` and `dim` the variable `subcomplexes[n_dim][k][dim]`
        stores the `dim`-simplices on the cover indexed by the `k` simplex of
        dimension `n_dim` in the nerve.
    zero_differentials : :obj:`list(list(list(Numpy Array)))`
        List storing the vertical differential matrices on the 0 page of the
        spectral sequence. For  integers `n_dim`, `k` and `dim` the variable
        `zero_differentials[n_dim][k][dim]` stores the `dim` differential of
        the complex on the cover indexed by the `k` simplex of dimension
        `n_dim` in the nerve.
    cycle_dimensions : :obj:`list(list(list(int)))`
        List storing the number of bars on each local persistent homology.
        Given two integers `n_dim` and `dim`, the variable
        `cycle_dimensions[n_dim][dim]` contains a list where each entry
        corresponds to an `n_dim` simplex in the nerve. For each such entry,
        we store the number of nontrivial persistent homology classes of
        dimension `dim` in the corresponding cover.
    Hom : :obj:`list(...(list(barcode_basis)))`
        Homology for each page of the spectral sequence. Given three integers
        which we denote `n_dim`, `nerv_spx` and `deg` we have that
        `Hom[0][n_dim][nerv_spx][deg]` stores a :obj:`barcode_basis` with the
        `deg`-persistent homology of the covering indexed by
        `nerve[n_dim][nerv_spx]`. All these store the homology on the `0` page
        of the spectral sequence. Additionally, for integers `k > 0`, `n_dim`
        and `deg`, we store in `Hom[k][n_dim][deg]` the :obj:`barcode_basis`
        for the homology on the `(deg, n_dim)` entry in the `k` page of the
        spectral sequence.
    Im : :obj:`list(...(list(barcode_basis)))`
        Image for each page of the spectral sequence. Given three integers
        which we denote `n_dim`, `nerv_spx` and `deg` we have that
        `Im[0][n_dim][nerv_spx][deg]` stores a :obj:`barcode_basis` for the
        image of the `deg+1`-differential of the covering indexed by
        `nerve[n_dim][nerv_spx]`. All these store the images on the `0` page
        of the spectral sequence. Additionally, for integers `k > 0`, `n_dim`
        and `deg`, we store in `Im[k][n_dim][deg]` the :obj:`barcode_basis` for
        the image on the `(deg, n_dim)` entry in the `k` page of the spectral
        sequence.
    PreIm : :obj:`list(...(list(Numpy Array)))`
        Preimages for each page of the spectral sequence. Given three integers
        which we denote `n_dim`, `nerv_spx` and `deg` we have that
        `PreIm[0][n_dim][nerv_spx][deg]` stores a :obj:`Numpy Array` for the
        Preimage of the `deg+1`-differential of the covering indexed by
        `nerve[n_dim][nerv_spx]`. Additionally, for integers `k > 0`, `n_dim`
        and `deg`, we store in `PreIm[k][n_dim][deg]` a :obj:`Numpy Array` for
        the preimages of the differential images in the `(deg, n_dim)` entry in
        the `k` page of the spectral sequence.
    tot_complex_reps : :obj:`list(list(*))`
        The asterisc `*` on the type can be either [] or
        :obj:`list(Numpy Array)`. This is used for storing complex
        representatives for the cycles.
    page_dim_matrix : :obj:`Numpy Array(no_pages+1, max_dim, no_columns)`
        Array storing the dimensions of the entries in each page. Notice that
        the order in which we store columns and rows differs from all the
        previous attributes.
    persistent_homology : :obj:`list(barcode_basis)`
        List storing the persistent homology generated by the spectral
        sequence. The `i` entry contains the `i` dimensional persistent
        homology.
    order_diagonal_basis : `list`
        This intends to store the original order of `persistent_homology`
        before applying the standard order.

    Notes
    -----
    The indexing on the 0 page is different from that of the next pages. This
    is because we do not want to store all the 0 page information on the same
    place.

    """
    def __init__(self, nerve, nerve_point_cloud, points_IN, max_dim,
                 max_r, no_pages, p):
        """Construction method
        """
        # dimensions of spectral sequence
        self.no_pages = no_pages
        self.no_rows = max_dim
        self.no_columns = len(nerve)
        self.max_r = max_r
        self.p = p
        # add nerve_point_cloud to spectral_sequence info
        self.nerve_point_cloud = nerve_point_cloud
        # add points IN to support Cech Differential
        self.points_IN = points_IN
        # add nerve and compute nerve differentials
        self.nerve = nerve
        self.nerve_differentials = complex_differentials(nerve, p)
        # list containing barcode bases for Hom, Im and PreIm
        # Hom and Im go through all pages, whereas
        # PreIm is only contained in the 0 page
        self.Hom = [[]]
        self.Im = [[]]
        self.PreIm = [[]]
        self.subcomplexes = []
        self.zero_differentials = []
        self.cycle_dimensions = []
        # vectors that translate local indices to global
        self.tot_complex_reps = []
        for n_dim in range(len(nerve)):
            self.Hom[0].append([])
            self.Im[0].append([])
            self.PreIm[0].append([])
            self.subcomplexes.append([])
            self.zero_differentials.append([])
            self.cycle_dimensions.append([])
            self.tot_complex_reps.append([])
            for deg in range(self.no_rows):
                self.tot_complex_reps[n_dim].append([])

        # make lists to store information in higher pages
        for k in range(1, no_pages):
            self.Hom.append([])
            self.Im.append([])
            self.PreIm.append([])
            for n_dim in range(self.no_columns):
                self.Hom[k].append([])
                self.Im[k].append([])
                self.PreIm[k].append([])
                for deg in range(self.no_rows):
                    self.Hom[k][n_dim].append([])
                    self.Im[k][n_dim].append([])
                    self.PreIm[k][n_dim].append([])

        # save space for dimension matrices for all pages
        # the order of variables is for printing the spectral sequence
        self.page_dim_matrix = np.zeros((no_pages+1, max_dim,
                                         self.no_columns)).astype(int)

        # define persistent homology and order of diagonal basis
        self.persistent_homology = []
        self.order_diagonal_basis = []

    ###########################################################################
    # add content to first page

    def add_output_first(self, output, n_dim):
        """Stores the 0 page data of `n_dim` column after it has been computed
        in parallel by `multiprocessing.pool`

        Parameters
        ----------
        output : :obj:`list`
            Result after using `multiprocessing.pool` on
            :meth:`permaviss.spectral_sequence.MV_spectral_seq.local_persistent_homology`
        n_dim : int
            Column of `0`-page whose data has been computed.

        """
        self.subcomplexes[n_dim] = [it[0] for it in output]
        self.zero_differentials[n_dim] = [it[1] for it in output]
        self.Hom[0][n_dim] = [it[2] for it in output]
        self.Im[0][n_dim] = [it[3] for it in output]
        self.PreIm[0][n_dim] = [it[4] for it in output]

        # Number of simplices in nerve[n_dim]
        if n_dim > 0:
            n_spx_number = np.size(self.nerve[n_dim], 0)
        else:
            n_spx_number = self.nerve[0]

        # add this to check that the level of intersection is not empty.
        # otherwise we might run out of index range
        if len(self.Hom[0][n_dim]) > 0:
            for deg in range(self.no_rows):
                no_cycles = 0
                # cumulative dimensions
                self.cycle_dimensions[n_dim].append(
                    np.zeros(n_spx_number+1).astype(int))
                for k in range(n_spx_number):
                    # Generate page dim matrix and local_coordinates info
                    cycles_in_cover = self.Hom[0][n_dim][k][deg].dim
                    no_cycles += cycles_in_cover
                    self.cycle_dimensions[n_dim][deg][k] = no_cycles
                # end for
                self.page_dim_matrix[1, deg, n_dim] = no_cycles
            # end for
        # end if

    ###########################################################################
    # add higher page contents

    def add_output_higher(self, Hom, Im, PreIm, end_n_dim, end_deg,
                          current_page):
        """Stores higher page data that has been computed along a sequence of
        consecutive differentials.

        The studied sequence of differentials ends in

            `(end_n_dim, end_deg)`

        coming from

            `(end_n_dim + current_page, end_deg - current_page + 1)`

        and continuing until reaching an integer `r > 0` such that either

                end_n_dim + r * current_page > self.no_columns

        or

                end_deg - r * current_page + 1 > 0

        Parameters
        ----------
        Hom : :obj:`list(barcode_basis)`
            Homology of a sequence of differentials in the spectral sequence.
            This is computed using
            :mod:`permaviss.persistence_algebra.module_persistence_homology`.
        Im : :obj:`list(barcode_basis)`
            Images of a sequence of differentials in the spectral sequence.
        PreIm : :obj:`list(Numpy Array)`
            Preimages of a sequence of differentials in the spectral sequence.
        end_n_dim : int
            Integer specifying the column position where the sequence of
            differentials ends.
        end_deg : int
            Integer specifying the row position where the sequence of
            differentials ends.
        current_page : int
            Current page of the spectral sequence.

        """
        n_dim = end_n_dim
        deg = end_deg
        for i, h in enumerate(Hom):
            self.PreIm[current_page][n_dim][deg] = PreIm[i]
            self.Hom[current_page][n_dim][deg] = h
            self.page_dim_matrix[current_page+1, deg, n_dim] = h.dim
            self.Im[current_page][n_dim][deg] = Im[i]
            deg = deg - current_page + 1
            n_dim += current_page

    ###########################################################################
    # zig_zag

    def zig_zag(self, n_dim, deg, current_page, lift_sum=True,
                initial_sum=[], store_reps=False):
        """Computes the image of the differential of the spectral sequence
        applied to a matrix of coordinates.

        Given an array of coordinates, computes the image under the
        differential in the current page. If no coordinates are given in
        initial_sum, we compute the image of all the generators on the position
        (deg, n_dim) of the `current_page`-page of the spectral sequence.

        Parameters
        ----------
        n_dim : int
            Column in page.
        deg : int
            Row in page
        current_page : int
            Current page in spectral sequence
        lift_sum : bool, default is True
            Whether we return a lifted sum on current_page or we return an
            expression on the 0 page instead.
        initial_sum : :obj:`list`, default is []
            Expression for which we want to compute zig-zag. Coordinates are
            stored on each column and are given in terms of the basis stored in
            `Hom` for the position `(deg, n_dim)` in `current_page -1`. If this
            is not given, we set it to be the identity of size equal to the
            dimension of the considered position on the spectral sequence.
        store_reps : bool,  default is False.
            Whether we want to store total complex representatives.

        Returns
        -------
        :obj:`Numpy Array`, if `lift_sum==False` then returns :obj:`list(dict)`
            Coordinates for images. Each column stores the image by the
            `current_page` differential applied to the corresponding column in
            `initial_sum`. These coordinates are given in terms of the basis
            stored in `Hom` for the position `(deg, n_dim)` for the page
            `current_page -1`. If `lift_sum` is set up to False, then this is a
            :obj:`list` where each entry stores a cochain. Each cochain is a
            dictionary :obj:`dict`, where entries are indexed by covering
            simplices, and these contain the local coordinates.

        """
        if len(initial_sum) == 0:
            initial_sum = np.identity(self.page_dim_matrix[current_page, deg,
                                                           n_dim])
        # end if

        # Save initial positions if we store reps
        if store_reps:
            start_n_dim = n_dim
            start_deg = deg
        # end if

        # Handle trivial arrows
        if self.page_dim_matrix[current_page, deg, n_dim] == 0:
            return np.array([])
        # end if

        # Load expression to zero page
        zero_coordinates, R = self.load_to_zero_page(initial_sum, n_dim, deg,
                                                     current_page)
        if store_reps:
            self.tot_complex_reps[start_n_dim][start_deg].append(
                zero_coordinates)
            if n_dim == 0:
                return
            # end if
        # end if

        # Compute zigzag for the zero_page expression
        zero_coordinates = self.cech_differential(zero_coordinates, n_dim, deg)

        n_dim -= 1
        for k in range(current_page - 1):
            zero_coordinates = self.lift_preimage(zero_coordinates,
                                                  R,  n_dim, deg)
            deg += 1
            if store_reps:
                self.tot_complex_reps[start_n_dim][start_deg].append(
                    zero_coordinates)
                if n_dim == 0:
                    return

            zero_coordinates = self.cech_differential(zero_coordinates,
                                                      n_dim, deg)
            n_dim -= 1
            # after this even though the class of zero_coordinates is zero
            # in current_page, this does not directly imply that these
            # can be lifted through the vertical differential. We adjust.
            if k < current_page - 2:
                self.liftable_0_class(zero_coordinates, R, n_dim, deg,
                                      current_page, store_reps,
                                      start_n_dim=start_n_dim,
                                      start_deg=start_deg)

        # Lift back to current_page
        if lift_sum:
            target_coordinates = self.lift_to_page(
                zero_coordinates, R,  n_dim, deg, current_page)
        else:
            target_coordinates = zero_coordinates

        return target_coordinates

    ###########################################################################
    # vert_image

    def vert_image(self, initial_sum, n_dim, deg):
        """Given an expression on zero page, computes the image under the
        vertical differentials.

        Parameters
        ----------
        initial_sum : :obj:`list(dict)`
            Coordinates of chains stored as a dictionary. Each dictionary entry
            corresponds to a `n_dim` simplex `s` on the covering nerve, and
            this corresponds to an intersection of covers `K_s`. On each such
            entry, there is a :obj:`Numpy Array` matrix storing the local
            coordinates for each chain. More precisely, the rows correspond to
            different cochains while the columns correspond `deg` simplices
            contained in `K_s`.

        """
        result = []
        for i, rep in enumerate(initial_sum):
            result.append({})
            for nerv_spx in iter(rep):
                image = multiply_mod_p(
                    self.zero_differentials[n_dim][nerv_spx][deg],
                    rep[nerv_spx], self.p)
                if np.any(image):
                    result[i][nerv_spx] = image
                # end if
            # end for
        # end for
        return result

    ###########################################################################
    # load_to_zero_page

    def load_to_zero_page(self, initial_sum, n_dim, deg, current_page):
        """Given an expression on the nth page of the spectral sequence, we
        pick a representative on the 0th page and return it.

        Parameters
        ----------
        initial_sum : :obj:`Numpy Array`
            Matrix storing the coordinates of different elements in terms of
            the basis stored in `Hom` at `(deg, n_dim)` and on the page
            `current_page`. Different element coordinates are on each column,
            while the basis corresponds to rows in the matrix.
        n_dim : int
            Column position on page.
        deg : int
            Row position on page
        current_page : int
            Current page number

        Returns
        -------
        target_coordinates : :obj:`list(dict)`
            List where each entry contains a dictionary storing cochains on
            zero page. Each dictionary entry corresponds to a `n_dim` simplex
            `s` on the covering nerve, and this contains the local coordinates
            for each cochain.
        R : :obj:`Numpy Array`
            1D array storing the radii of birth for each cochain stored in
            `target_coordinates`.

        """
        # Load sum to 1st page
        aux_sum = initial_sum
        for k in range(current_page-1, 0, -1):
            aux_sum = multiply_mod_p(self.Hom[k][n_dim][deg].coordinates,
                                     aux_sum, self.p)
        # end for

        # Load sum from 1 to 0 page
        # Transform coordinates to dictionary data types
        target_coordinates = []
        for i, coord in enumerate(aux_sum.T):
            target_coordinates.append({})
            for nerve_spx in range(len(self.cycle_dimensions[n_dim][deg])-1):
                local_coordinates = coord[
                    self.cycle_dimensions[n_dim][deg][nerve_spx-1]:
                    self.cycle_dimensions[n_dim][deg][nerve_spx]]

                # If the local coordinates are nontrivial, add them to
                # dictionary.
                if len(local_coordinates) > 0 and np.any(local_coordinates):
                    target_coordinates[i][nerve_spx] = multiply_mod_p(
                        self.Hom[0][n_dim][nerve_spx][deg].coordinates,
                        local_coordinates, self.p)
                # end if
            # end for
        # end for
        # Get initial radius of barcodes
        R = np.zeros(int(self.page_dim_matrix[current_page, deg, n_dim]))
        if current_page == 1:
            for i, coord in enumerate(initial_sum.T):
                for nerve_spx in range(len(self.cycle_dimensions[n_dim][
                        deg]) - 1):
                    local_coordinates = coord[self.cycle_dimensions[n_dim][
                        deg][nerve_spx - 1]: self.cycle_dimensions[
                            n_dim][deg][nerve_spx]]
                    # Take the maximum of birth radius over all nontrivial
                    # local coordinates.
                    if len(local_coordinates) > 0 and np.any(
                            local_coordinates):
                        R[i] = max(R[i], self.Hom[0][n_dim][nerve_spx][
                            deg].birth_radius(local_coordinates))
                    # end if
                # end for
            # end for
        else:
            for i, coord in enumerate(initial_sum.T):
                R[i] = self.Hom[current_page - 1][n_dim][
                    deg].birth_radius(coord)
            # end for
        # end else

        return target_coordinates, R

    ###########################################################################
    # cech_differential

    def cech_differential(self, start_coordinates, n_dim, deg):
        """Computes the Cech Differential of various cochains.

        That is, given a few cochains on the position `(deg, n_dim)` of the
        `0` page, we compute the Cech differential. This leads to cochains on
        `(deg, n_dim - 1)`.

        Parameters
        ----------
        start_coordinates : :obj:`list(dict)`
            List where each entry contains a dictionary storing the
            coordinates of a cochain in `(n_dim, deg)`. Each dictionary key
            corresponds to a `n_dim` simplex `s` on the covering nerve, and
            this contains the local coordinates for each cochain.
        n_dim : int
            Column on page
        deg : int
            Row on page

        Returns
        -------
        target_coordinates : :obj:`list(dict)`
            Cech differential images of `start_coordinates`. These are stored
            as a list of cochains in `(n_dim-1, deg)` with the same format as
            `start_coordinates`

        """

        target_coordinates = []
        deg_sign = (-1)**deg
        # go through all dictionaries in input
        for i, coordinates in enumerate(start_coordinates):
            cech_image = {}
            # look at each nerve simplex on each dictionary
            for nerve_spx_index in iter(coordinates):
                points_IN = np.copy(self.points_IN[n_dim][nerve_spx_index])
                # Iterate over simplices in boundary of nerve simplex
                nerv_boundary_indx = np.nonzero(self.nerve_differentials[
                    n_dim][:, nerve_spx_index])[0]
                nerv_boundary_coeff = self.nerve_differentials[n_dim][
                    :, nerve_spx_index][nerv_boundary_indx]
                # go through all faces for each simplex in nerve
                for nerve_face_index, nerve_coeff in zip(
                        nerv_boundary_indx, nerv_boundary_coeff):
                    if deg == 0:
                        # inclusions for points
                        # Preallocate space
                        if nerve_face_index not in cech_image:
                            cech_image[nerve_face_index] = np.zeros(
                                self.subcomplexes[n_dim - 1][
                                    nerve_face_index][0])
                        # end if
                        for point_idx, point_coeff in enumerate(coordinates[
                                nerve_spx_index]):
                            if point_coeff != 0:
                                face_point_idx = np.argmax(self.points_IN[
                                    n_dim-1][nerve_face_index] == points_IN[
                                        point_idx])
                                cech_image[nerve_face_index][
                                    face_point_idx
                                    ] += nerve_coeff * point_coeff * deg_sign
                                cech_image[nerve_face_index][
                                    face_point_idx] %= self.p
                            # end if
                        # end for
                    else:
                        # inclusions for edges, 2-simplices and higher
                        # simplices. Preallocate space as well.
                        if nerve_face_index not in cech_image:
                            cech_image[nerve_face_index] = np.zeros(len(
                                self.subcomplexes[n_dim - 1][nerve_face_index][
                                    deg]))
                        # end if
                        # Iterate over nontrivial local simplices in domain
                        spx_indices = np.nonzero(coordinates[
                            nerve_spx_index])[0]
                        spx_coefficients = coordinates[nerve_spx_index][
                            spx_indices]
                        for spx_index, spx_coeff in zip(
                                spx_indices, spx_coefficients):
                            # Obtain IN for vertices of simplex
                            vertices_spx = points_IN[self.subcomplexes[n_dim][
                                nerve_spx_index][deg][spx_index]]
                            # Iterate over simplices in range to see which
                            # one has vertices_spx as vertices.
                            for im_indx, im_spx in enumerate(
                                    self.subcomplexes[n_dim-1][
                                        nerve_face_index][deg]):
                                vertices_face = self.points_IN[n_dim-1][
                                    nerve_face_index][im_spx.astype(int)]
                                # When the vertices coincide, break the loop
                                if len(np.intersect1d(
                                        vertices_spx,
                                        vertices_face)) == deg + 1:
                                    cech_image[nerve_face_index][im_indx] += \
                                        spx_coeff * nerve_coeff * deg_sign
                                    cech_image[nerve_face_index][im_indx] %= \
                                        self.p
                                    break
                                # end if
                            # end for
                        # end for
                    # end else
                # end for
            # end for
            target_coordinates.append(cech_image)
        # end for

        return target_coordinates

    ###########################################################################
    # lift_to_page

    def lift_to_page(self, start_coordinates, R,  n_dim, deg, target_page):
        """Lifts some zero page element to a target page.

        This is, in a way, opposite to
        :meth:`permaviss.spectral_sequence.spectral_sequence_class.load_to_zero_page`.
        More precisely, given a few cochains in `start_coordinates`, we compute
        their respective classes on `target_page`.

        Parameters
        ----------
        start_coordinates : :obj:`list(dict)`
            List where each entry contains a dictionary storing cochains on
            zero page. Each dictionary entry corresponds to a `n_dim` simplex
            `s` on the covering nerve, and this contains the local coordinates
            for each cochain.
        R : :obj:`Numpy Array`
            1D array storing the radii of birth for each cochain stored in
            `start_coordinates`.
        n_dim : int
            Column on page
        deg : int
            Row on page
        target_page : int
            Page to which we want to lift `start_coordinates`.

        Returns
        -------
        target_coordinates : :obj:`Numpy Array`
            Matrix storing the coordinates for each class corresponding to each
            cochain in `start_coordinates`. Rows correspond to the number of
            classes, while the columns correspond to the dimension of
            `Hom[n_dim][deg][target_page]`.

        """
        # Lift from 0 to 1 page
        # Transform from dictionary to np.array
        target_coordinates = np.zeros((len(start_coordinates),
                                       self.page_dim_matrix[1, deg, n_dim]))
        for i, coord in enumerate(start_coordinates):
            for nerve_spx_index in iter(coord):
                # solve (Im|Hom) locally
                # This is with respect to the active coordinates at R[i]
                Hom_dim = self.Hom[0][n_dim][nerve_spx_index][deg].dim
                if Hom_dim > 0:
                    Hom = self.Hom[0][n_dim][nerve_spx_index][
                        deg].active_domain(R[i])
                Im_dim = self.Im[0][n_dim][nerve_spx_index][deg].dim
                if Hom_dim > 0 and len(Hom) > 0:
                    if Im_dim > 0:
                        Im = self.Im[0][n_dim][nerve_spx_index][
                            deg].active_domain(R[i])
                        Im_dim = np.size(Im, 1)
                        Im_Hom = np.append(Im, Hom, axis=1)

                    else:
                        Im_Hom = Hom
                    # end else

                    active_coordinates = solve_mod_p(
                        Im_Hom, coord[nerve_spx_index], self.p)[Im_dim:]
                    # Save resulting coordinates as 1 page coordinates
                    local_lifted_coordinates = np.zeros(self.cycle_dimensions[
                        n_dim][deg][nerve_spx_index] - self.cycle_dimensions[
                            n_dim][deg][nerve_spx_index-1])
                    local_active = self.Hom[0][n_dim][nerve_spx_index][
                        deg].active(R[i])
                    local_lifted_coordinates[local_active] = active_coordinates
                    target_coordinates[i, self.cycle_dimensions[n_dim][deg][
                        nerve_spx_index-1]:self.cycle_dimensions[n_dim][deg][
                            nerve_spx_index]] = local_lifted_coordinates
                # end if
            # end for
        # end for

        # Lift from 1 to target page
        for k in range(1, target_page):
            prev_target_coordinates = target_coordinates
            target_coordinates = np.zeros((
                len(start_coordinates), self.page_dim_matrix[k+1, deg, n_dim]))
            for i, coord in enumerate(prev_target_coordinates):
                # solve (Im|Hom) locally
                Hom_dim = self.Hom[k][n_dim][deg].dim
                if Hom_dim > 0:
                    Hom = self.Hom[k][n_dim][deg].active_coordinates(R[i])

                Im_dim = self.Im[k][n_dim][deg].dim
                if Hom_dim > 0 and len(Hom) > 0:
                    if Im_dim > 0:
                        Im = self.Im[k][n_dim][deg].active_coordinates(R[i])
                        Im_dim = np.size(Im, 1)
                        Im_Hom = np.append(Im, Hom, axis=1)
                    else:
                        Im_Hom = Hom
                    # end else
                    lifted_coordinates = np.zeros(self.Hom[k][n_dim][deg].dim)
                    if np.size(Im_Hom, 1) > 0:
                        prev_active_coordinates = prev_target_coordinates[i][
                            self.Hom[k][n_dim][deg].prev_basis.active(R[i])]
                        if np.any(prev_active_coordinates):
                            active_coordinates = solve_mod_p(
                                Im_Hom, prev_active_coordinates,
                                self.p)[Im_dim:]
                            lifted_coordinates[self.Hom[k][n_dim][deg].active(
                                R[i])] = active_coordinates
                        # end if
                    # end if
                    target_coordinates[i] = lifted_coordinates
                # end if
            # end for
        # end for

        return target_coordinates

    ###########################################################################
    # lift_preimage

    def lift_preimage(self,  start_coordinates, R, n_dim, deg, sign=-1):
        """Given a zero page element, we lift it using the vertical
        differentials.

        Additionally, we multiply the lift by -1 mod self.p
        In this method we assume that this can be done.

        Parameters
        ----------
        start_coordinates : :obj:`list(dict)`
            These are the cochains on the `(deg, n_dim)` position that we want
            to lift through the vertical differential. List containing cochains
            on `(deg, n_dim)`. Each entry contains the coefficients of a
            cochain as a dictionary indexed by `n_dim` simplices on the nerve
            of the cover.
        R : :obj:`Numpy Array`
            1D array storing the radii of birth for each cochain stored in
            `start_coordinates`.
        n_dim : int
            Column on page
        deg : int
            Row on page

        Returns
        -------
        lifted_coordinates : :obj:`list(dict)`
            These are the cochains on the `(deg+1, n_dim)` position that we
            want to lift through the vertical differential. List containing
            cochains on `(deg+1, n_dim)`. Each entry contains the coefficients
            of a cochain as a dictionary indexed by `n_dim` simplices on the
            nerve of the cover.

        """
        lifted_coordinates = []
        for i, coord in enumerate(start_coordinates):
            lifted_coordinates.append({})
            for nerve_spx_index in iter(coord):
                # Only lift nontrivial expressions
                if np.any(coord[nerve_spx_index]):
                    Im = self.Im[0][n_dim][nerve_spx_index][
                        deg].active_domain(R[i])
                    solution = solve_mod_p(Im, coord[nerve_spx_index], self.p)
                    local_lift = multiply_mod_p(
                        self.PreIm[0][n_dim][nerve_spx_index][deg+1][
                            :, self.Im[0][n_dim][nerve_spx_index][
                                deg].active(R[i])],
                        solution, self.p)
                    # lifted coordinates in terms of all coordinates
                    lifted_coordinates[i][
                        nerve_spx_index] = sign * local_lift % self.p
                # end if
            # end for
        # end for

        return lifted_coordinates

    ###########################################################################
    # extension

    def extension(self, start_n_dim, start_deg):
        """Generate extension block matrices for given n_dim and deg

        Parameters
        ----------
        start_n_dim : int
            Column on page where we want to compute the extension coefficients.

        start_deg : int
            Row on page where we want to compute the extension coefficients.

        Returns
        -------
        extension : :obj:`list(Numpy Array)`
            Extension matrices for the basis contained at `(start_deg,
            start_n_dim)` of the infinity page. More precisely, given an
            integer `ext_dim`, the entry `extension[ext_dim]` stores a
            :obj:`Numpy Array` matrix where the `i` column corresponds to the
            `i` generator in the position (start_deg, start_n_dim) from the
            infinity page. Each such column contains the extension
            coefficients in terms of the generators from the
            `(deg + ext_deg, n_dim - ext_deg)` entry on the infinity page.
            Recall that each slope -1 diagonal on the infinity page corresponds
            to a `broken basis` for the persistent homology. Then, each of the
            matrices in `extension_matrices` corresponds to a block in the
            `broken differentials`. Notice that the order in which we store
            columns and rows differs from all the previous attributes except
            page_dim_matrix.

        """
        extensions = []
        death_R = self.Hom[self.no_pages-1][start_n_dim][
            start_deg].barcode[:, 1]
        dim_domain = self.Hom[self.no_pages-1][start_n_dim][start_deg].dim
        not_infty = death_R < self.max_r
        death_R = death_R[not_infty]
        not_infty = np.sort(np.nonzero(not_infty)[0])
        dim_ext = len(not_infty)
        # Take only the representatives that are not infinity
        total_representatives = self.tot_complex_reps[start_n_dim][start_deg]
        representatives = []
        for i, zero_coord in enumerate(total_representatives):
            representatives.append([
                copy_dictionary(zero_coord[i]) for i in not_infty])

        # Start in specified entry, and go backwards
        ext_deg = 0
        deg = start_deg
        for n_dim in range(start_n_dim, -1, -1):
            # find and store extension coefficients
            extension = np.zeros((self.page_dim_matrix[
                self.no_pages-1][deg][n_dim], dim_domain))
            if len(not_infty) > 0:
                extension_coefficients = self.lift_to_page(
                    representatives[ext_deg], death_R, n_dim, deg,
                    self.no_pages-1)
                extension[:, not_infty] = extension_coefficients.T
                # change total chain representatives accordingly
                current_representatives = self.tot_complex_reps[n_dim][deg]
                for i in range(dim_ext):
                    # subtract sum of coefficients along all total complex,
                    # starting at n_dim, deg.
                    for k, rep in enumerate(current_representatives):
                        substractor = add_dictionaries(
                            extension_coefficients[i], rep, self.p)
                        representatives[k+ext_deg][i] = add_dictionaries(
                            [1, -1],
                            [representatives[k+ext_deg][i], substractor],
                            self.p)
                    # end for
                # end for
                # subtract images to representatives, so that we can lift the
                # vertical differentials.
                if n_dim > 0:
                    # subtract images of differentials so that we can lift
                    self.liftable_0_class(representatives[ext_deg], death_R,
                                          n_dim, deg, self.no_pages - 1, False)
                    # lift through vertical differential the last nontrivial
                    # entry in total complex diagonal.
                    preimages_sums = self.lift_preimage(representatives[
                        ext_deg], death_R, n_dim, deg)
                    representatives[ext_deg] = {}
                    # compute the cech differential for preimage_sums
                    cech_images = self.cech_differential(preimages_sums, n_dim,
                                                         deg + 1)
                    # subtract cech differential of preimage
                    for i in range(dim_ext):
                        representatives[ext_deg + 1][i] = add_dictionaries([
                            1, 1], [representatives[ext_deg + 1][i],
                                    cech_images[i]], self.p)
                    # end for
                # end if
            # end if
            extensions.append(extension)
            deg += 1
            ext_deg += 1
        # end for

        return extensions

    ###########################################################################
    # liftable_0_class

    def liftable_0_class(self, start_coordinates, R, n_dim, deg, target_page,
                         store_reps, start_n_dim=None, start_deg=None):
        """ Given cochains on the `0`-page, and assuming that they are zero
        when lifted to `target_page`, we get representatives that are
        equivalent on `target_page - 1` and also lift through the vertical
        differential.

        Parameters
        ----------
        start_coordinates : :obj:`list(dict)`
            These are the cochains on the `(deg, n_dim)` position that we want
            to lift through the vertical differential. List containing cochains
            on `(deg, n_dim)`. Each entry contains the coefficients of a
            cochain as a dictionary indexed by `n_dim` simplices on the nerve
            of the cover. This is modified so that it can be lifted.
        R : :obj:`Numpy Array`
            1D array storing the radii of birth for each cochain stored in
            `start_coordinates`
        n_dim : int
            Column on page
        deg : int
            Row on page
        target_page : int
            Page of spectral sequence at which the given classes vanish.

        Returns
        -------
        liftable_coordinates : :obj:`list(dict)`
            Cochains on the `(deg, n_dim)` position whose classes on
            `target_page` are equivalent to those of `start_coordinates`.
            Furthermore, these can also be lifted through vertical
            differentials.

        """
        dim_reps = len(start_coordinates)
        for page in range(target_page - 1, 0, -1):
            image_coefficients = self.image_coordinates(start_coordinates, R,
                                                        n_dim, deg, page)
            # If image in position is nontrivial
            if len(image_coefficients) > 0 and np.any(image_coefficients):
                preimages_sums = multiply_mod_p(
                    image_coefficients,
                    self.PreIm[page][n_dim+page][deg-page+1].T,
                    self.p)
                # subtract images to zero page sums
                if np.any(preimages_sums):
                    images = self.zig_zag(
                        n_dim+page, deg-page+1, page,
                        lift_sum=False,
                        initial_sum=preimages_sums.T)
                    preimage_zeros, _ = self.load_to_zero_page(
                        preimages_sums.T, n_dim+page, deg-page+1, page)
                    tot_complex_reps = self.tot_complex_reps[
                        start_n_dim][start_deg][deg - page + 1 - start_deg]
                    for i in range(dim_reps):
                        start_coordinates[i] = add_dictionaries(
                            [1, -1], [start_coordinates[i], images[i]],
                            self.p)
                        if store_reps:
                            tot_complex_reps[i] = add_dictionaries(
                                [1, -1],
                                [tot_complex_reps[i], preimage_zeros[i]],
                                self.p)

                    # end for
                # end if
            # end if
        # end for

    ###########################################################################
    # image_coordinates

    def image_coordinates(self, start_coordinates, R, n_dim, deg,
                          target_page):
        """Given cochains on the `0`-page, we compute the coordinates of their
        persistent Homology classes in terms of the basis in
        `Im[target_page][n_dim][deg]`

        Parameters
        ----------
        start_coordinates : :obj:`list(dict)`
            These are the cochains on the `(deg, n_dim)` position that we want
            to lift through the vertical differential. List containing cochains
            on `(deg, n_dim)`. Each entry contains the coefficients of a
            cochain as a dictionary indexed by `n_dim` simplices on the nerve
            of the cover.
        R : :obj:`Numpy Array`
            1D array storing the radii of birth for each cochain stored in
            `start_coordinates`
        n_dim : int
            Column on page
        deg : int
            Row on page
        target_page : int
            Page of spectral sequence at which we want to solve the image
            equation.

        Returns
        -------
        image_coordinates : :obj:`Numpy Array`
            Start_coordinates classes in target_page written in terms of
            image of differential. Rows correspond to different vectors
            and columns correspond to the dimension of the image.

        Raises
        ------
        ValueError
            If the classes of the cochains are not contained in the images.

        """
        # Lift start_coordinates from 0 to target_page
        lifted_sum = self.lift_to_page(start_coordinates, R, n_dim, deg,
                                       target_page)

        # If image is trivial, return
        Im_dim = self.Im[target_page][n_dim][deg].dim
        if Im_dim == 0:
            return []

        # Compute each coordinate in terms of images
        image_coordinates = np.zeros((len(start_coordinates), Im_dim))
        for i, rad in enumerate(R):
            Im = self.Im[target_page][n_dim][deg].active_coordinates(rad)
            active_lifted_sum = lifted_sum[i][self.Im[target_page][n_dim][
                deg].prev_basis.active(rad)]
            if len(active_lifted_sum) > 0:
                image_coordinates[i, self.Im[target_page][n_dim][
                    deg].active(rad)] = solve_mod_p(
                        Im, active_lifted_sum, self.p)

        return image_coordinates
    # end def
# end spectral_sequence class

###############################################################################
# add_dictionaries
#
# This mainly supports extension. Does what its name says.
# That is, entries with the same key are added.


def add_dictionaries(coefficients, representatives, p):
    """ Computes a dictionary that is the linear combination of `coefficients`
    on `representatives`

    Parameters
    ----------
    coefficients : :obj:`Numpy Array`
        1D array with the same number of elements as `representatives`. Each
        entry is an integer mod p.
    representatives : :obj:`list(dict)`
        List where each entry is a dictionary. The keys on each dictionary are
        integers, and these might coincide with dictionaries on other entries.
    p : int(prime)

    Returns
    -------
    rep_sum : :obj:`dict`
        Result of adding the dictionaries on `representatives` with
        `coefficients`.

    Example
    -------
        >>> import numpy as np
        >>> p=5
        >>> coefficients = np.array([1,2,3])
        >>> representatives = [
        ... {0:np.array([1,3]), 3:np.array([0,0,1])},
        ... {0:np.array([4,3]),2:np.array([4,5])},
        ... {3:np.array([0,4,0])}]
        >>> add_dictionaries(coefficients, representatives, p)
        {0: array([4, 4]), 3: array([0, 2, 1]), 2: array([3, 0])}


    """
    rep_sum = {}
    for i, rep in enumerate(representatives):
        for spx_idx in iter(rep):
            if spx_idx not in rep_sum:
                rep_sum[spx_idx] = (coefficients[i] * rep[spx_idx]) % p
            else:
                rep_sum[spx_idx] = (rep_sum[spx_idx] + coefficients[i] * rep[
                    spx_idx]) % p
            # end else
        # end for
    # end for
    # Find simplices where expression is zero
    zero_simplices = []
    for spx_idx in iter(rep_sum):
        if not np.any(rep_sum[spx_idx]):
            zero_simplices.append(spx_idx)
        # end if
    # end for
    # If an entry is zero,  delete it
    for spx_idx in zero_simplices:
        del rep_sum[spx_idx]
    # end for

    return rep_sum


###############################################################################
# copy_dictionary
#

def copy_dictionary(original):
    """Copies a dictionary where each entry is a :obj:`Numpy Array`
    """
    copy = {}
    for spx_idx in iter(original):
        copy[spx_idx] = np.copy(original[spx_idx])

    return copy



"""
    differentials.py

"""
import numpy as np


def complex_differentials(C, p):
    """Given a simplicial complex C, it returns a list D with its
    differentials mod p

    Parameters
    ----------
    C : :obj:`list(int, Numpy Array, ...)`
        The `0` entry stores the number of vertices. For a higher entry `i`,
        `C[i]` stores a :obj:`Numpy Array` matrix with the `i` simplices
        from `C`.
    D : :obj:`list(Numpy Array)`
        List of differentials of `C`. The `i` entry contains a :obj:`Numpy
        Array` of the `i` differential for the simplicial complex.

    """
    dim = len(C) - 1  # get the dimension of the simplicial complex
    D = []
    for d in range(dim + 1):
        D.append([])

    D[0] = C[0]  # number of points
    # create 1D boundary
    D[1] = np.zeros((C[0], np.size(C[1], 0)))
    for i, c in enumerate(C[1]):
        D[1][int(c[0])][i] = 1
        D[1][int(c[1])][i] = p - 1

    # We proceed to compute all the differentials for all dimensions
    for d in range(2, dim + 1):
        """
            Go through all the d-simplexes of C, computing their images
            in (d-1)-simplex basis
        """
        D[d] = np.zeros((np.size(C[d-1], 0), np.size(C[d], 0)))
        for i, s in enumerate(C[d]):
            im = []  # initialize image of s
            signs = []  # also initialize signs of faces
            for j, c in enumerate(C[d - 1]):
                if len(np.intersect1d(c, s)) == d:
                    face_idx = np.setdiff1d(s, c)[0]
                    im.append(j)
                    signs.append((-1) ** np.searchsorted(s, face_idx))

            for j, v in enumerate(im):
                D[d][v][i] = signs[j] % p

    return D
